/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * DBConnPool.java
 *
 * Created on 30 August 2008, 04:50
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package net.infantai.dbcp;

import java.sql.*;
import java.util.*;

/**
 * This class provides the functionality to create and maintain a pool of Database Connection objects.
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class DBConnPool
{
    
    private static final int MAX_CONNECTIONS = 150;
    
//    private static DBConnPool dbcp = null;

    private String dbDriverClassName = null;
    private String dbServer = null;
    private String dbServerPort = null;
    private String dbName = null;
    private String dbUsername = null;
    private String dbPassword = null;
    private String dbConnectionString = null;

    private int initPoolSize = 2;
    private int maxPoolSize = 150; // Default
    private int maxWaitTime = 15000; // Milliseconds
    private int checkInterval = 50; // Milliseconds

    private Map<String, Connection> dbConnPool;
    private Map<String, Connection> connectionsInUse;
    private Queue<String> availableConnectionKeys;
    private Map<String, Long> lastUsedTimestamps;

    private DBConnParam dbParam = null;

    private String poolIdentifier = null;
    private static final Map<String, DBConnPool> poolOfPools = new LinkedHashMap<String, DBConnPool>();
    
    private final Object poolLock = new Object();
    
    /**
     * Creates a new instance of DBConnPool.
     */
    private DBConnPool(String poolIdentifier) throws DBCPException
    {     
        dbParam = DBConnParam.getInstance(poolIdentifier);
        
//        if (dbParam == null || dbParam.getDBConnectionString() == null || dbParam.getDBUsername() == null || dbParam.getDBPassword() == null)
//        {
//            throw new DBCPException("Parameters Not Set ! " + poolIdentifier);
//        }
        
        dbDriverClassName = dbParam.getDBDriverClassName();
        dbServer = dbParam.getDBServer();
        dbServerPort = dbParam.getDBServerPort();
        dbName = dbParam.getDBName();
        dbUsername = dbParam.getDBUsername();
        dbPassword = dbParam.getDBPassword();
        dbConnectionString = dbParam.getDBConnectionString();
        
        if (dbServerPort != null && dbServer != null && dbName != null && dbUsername != null && dbPassword != null)
        {
            dbConnectionString = dbConnectionString.replace("$dbserverport", dbServerPort);
            dbConnectionString = dbConnectionString.replace("$dbserver", dbServer);
            dbConnectionString = dbConnectionString.replace("$dbname", dbName);
            dbConnectionString = dbConnectionString.replace("$dbusername", dbUsername);
            dbConnectionString = dbConnectionString.replace("$dbpassword", dbPassword);        
        }
        
        if (maxPoolSize < MAX_CONNECTIONS)
        {
            maxPoolSize = dbParam.getMaxPoolSize();
        }
        else
        {
            maxPoolSize = MAX_CONNECTIONS;
        }
        maxWaitTime = dbParam.getMaxWait();
        this.poolIdentifier = poolIdentifier;
        try
        {
            initPool();
        }
        catch (SQLException sqlex)
        {
            throw new DBCPException("Connection Failed ! " + poolIdentifier);
        }
    }
    
    public static DBConnPool getInstance(String poolIdentifier) throws DBCPException
    {
        if (poolOfPools.get(poolIdentifier) == null)
        {
            DBConnPool dbcp = new DBConnPool(poolIdentifier);
            poolOfPools.put(poolIdentifier, dbcp);
        }
        return poolOfPools.get(poolIdentifier);
    }
    
    private void initPool() throws SQLException
    {
        dbConnPool = new LinkedHashMap<String, Connection>();
        connectionsInUse = new LinkedHashMap<String, Connection>();
        lastUsedTimestamps = new LinkedHashMap<String, Long>();
        availableConnectionKeys = new ArrayDeque<String>();
        loadConnections(initPoolSize);
        System.out.println("Initialized Database Connection Pool with " + dbConnPool.size() + " Connection object(s).");
    }
    
    private void destroyPool(String poolIdentifier)
    {
        poolOfPools.remove(poolIdentifier);
    }

    private void loadConnections(int count) throws SQLException
    {
        String uniqueKey = null;
        long ctms = System.currentTimeMillis();
        Random random = new Random(ctms);
        for (int i = 0; i < count; ++i)
        {
            Connection dbConn = getConnection();
            if (dbConn != null)
            {
                String strCnNumber = null;
                if (i < 9)
                {
                    strCnNumber = "0" + (i + 1);
                }
                else
                {
                    strCnNumber = Integer.toString((i + 1));
                }
                
                synchronized (poolLock)
                {
                    uniqueKey = "Infant*DBConn~" + strCnNumber + "~" + random.nextInt((initPoolSize * 1000)) + "*" + ctms;
                    dbConnPool.put(uniqueKey, dbConn);
                    availableConnectionKeys.add(uniqueKey);                
                }
            }
        }
    }
    
    private Connection getConnection() throws SQLException
    {
        Connection conn = null;
        try
        {
            Class.forName(dbDriverClassName);
            conn = DriverManager.getConnection(dbConnectionString, dbUsername, dbPassword);
        }
        catch (ClassNotFoundException cnfe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + cnfe.getMessage());
            cnfe.printStackTrace();
        }
        catch (SQLException ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "ERROR: Unable to connect to database. [SEVERITY LEVEL: CRITICAL]");
            destroyPool(this.poolIdentifier);
            throw ex;
        }
        return conn;
    }
    
    /**
     * Gets a Connection object from the pool.
     */
    public DBConn getDBConnection() throws InterruptedException, SQLException
    {
        Connection conn = null;
        DBConn dbConn = null;

        if (dbConnPool.size() == 0)
        {
            if (connectionsInUse.size() == maxPoolSize)
            {
                int wtimeelapsed = 0;
                while (wtimeelapsed < maxWaitTime)
                {
                    Thread.sleep(checkInterval);
                    wtimeelapsed += checkInterval;
                    if (dbConnPool.size() > 0)
                    {
                        break;
                    }
                }            
            }
        }

        synchronized (poolLock)
        {
            String connectionKey = availableConnectionKeys.poll();
            if (connectionKey != null)
            {
                conn = dbConnPool.remove(connectionKey);
                if (conn != null)
                {
                    dbConn = new DBConn();
                    dbConn.setConnection(conn);
                    dbConn.setUniqueKey(connectionKey);
                    connectionsInUse.put(connectionKey, conn);
                }
            }
            
            if (dbConnPool.size() == 0)
            {
                if (getSize() < maxPoolSize)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Warning ! Pool empty, adding 2 more Connections.");
                    loadConnections(2);
                }
                else
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Warning ! Pool empty, can't add more connections !");
                }
            }            
        }        
        
        return dbConn;
    }
    
    /**
     * Returns the specified connection back to the pool.
     */
    public void freeDBConnection(DBConn dbConn) throws SQLException
    {
        if (dbConn != null)
        {
            try
            {
                String uniqueKey = dbConn.getUniqueKey();
                Connection conn = dbConn.getConnection();
                if (conn != null && !conn.isClosed())
                {
                    synchronized (poolLock)
                    {
                        if (dbConnPool.size() < maxPoolSize)
                        {
                            connectionsInUse.remove(uniqueKey);
                            dbConnPool.put(uniqueKey, conn);
                            availableConnectionKeys.add(uniqueKey);
                        }                    
                    }
                }
                else
                {
                    if (uniqueKey != null)
                    {
                        System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + "Warning: Replacing NULL Database Connection => " + uniqueKey);
                        conn = getConnection();
                        synchronized (poolLock)
                        {
                            connectionsInUse.remove(uniqueKey);
                            dbConnPool.put(uniqueKey, conn);     
                            availableConnectionKeys.add(uniqueKey);
                        }
                    }
                }
                lastUsedTimestamps.put(uniqueKey, new Long(System.currentTimeMillis()));
            }
            catch (SQLException sqlex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + sqlex.getMessage());
                throw sqlex;
            }
        }
    }
    
    /**
     * Returns the number of Connection objects currently available in the pool.
     */
    public int getAvailableCount()
    {
        return dbConnPool.size();
    }
    
    public int getSize()
    {
        return dbConnPool.size() + connectionsInUse.size();
    }
    
}
