/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * DBCPWrapper.java
 *
 * Created on 05 May 2008, 17:51
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */
package net.infantai.dbcp.secure;

import java.sql.*;
import java.util.GregorianCalendar;
import java.util.concurrent.atomic.AtomicReference;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class DBCPWrapper
{

    private DBConnPool dbcp = null;
    private DBConn dbConn = null;
    private Connection conn = null;

    private static final String ERR_POOL_MAXED_OUT = "Error getting connection from pool, possibly caused by too many active connections.";

    /**
     * Creates a new instance of DBCPWrapper.
     *
     * @param poolIdentifier - The PoolIdentifier.
     */
    public DBCPWrapper(String poolIdentifier) throws DBCPException, SQLException
    {
        dbcp = DBConnPool.getInstance(poolIdentifier);
        dbConn = dbcp.getDBConnection(); // Get a connection from the pool.
        if (dbConn == null)
        {
            throw new DBCPException("Error getting connection from pool, possibly caused by too many active connections or loss of connectivity with Database Server. (Size: " + dbcp.getSize() + ") @ " + new java.util.Date().toString() + " [ " + poolIdentifier + " ]");
        }
        conn = dbConn.getConnection();
    }

    /*
    public ResultSet executeSelect(String sqlSelect, AtomicReference<Statement> outStmt) throws SQLException, DBCPException
    {
        if (conn != null)
        {
            Statement statement = conn.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);
            outStmt.set(statement);
            return statement.executeQuery(sqlSelect);
        }
        else
        {
            throw new DBCPException(ERR_POOL_MAXED_OUT);
        }
    }

    public int executeUpdate(String sqlUpdate) throws SQLException, DBCPException
    {
        int r = -1;
        if (conn != null)
        {
            Statement statement = conn.createStatement();
            r = statement.executeUpdate(sqlUpdate);
            statement.close();
        }
        else
        {
            throw new DBCPException(ERR_POOL_MAXED_OUT);
        }
        return r;
    }

    public ResultSet executeInsertGetGeneratedKeys(String sqlInsert) throws SQLException, DBCPException
    {
        if (conn != null)
        {
            Statement statement = conn.createStatement();
            statement.executeUpdate(sqlInsert, Statement.RETURN_GENERATED_KEYS);
            return statement.getGeneratedKeys();
        }
        else
        {
            throw new DBCPException(ERR_POOL_MAXED_OUT);
        }
    }
    */

    public PreparedStatement prepareStatement(String sqlStatement) throws SQLException, DBCPException
    {
        if (conn != null)
        {
            PreparedStatement pstmt = conn.prepareStatement(sqlStatement);
            return pstmt;
        }
        else
        {
            throw new DBCPException(ERR_POOL_MAXED_OUT);
        }
    }

    public CallableStatement prepareCall(String sqlStatement) throws SQLException, DBCPException
    {
        if (conn != null)
        {
            CallableStatement cstmt = conn.prepareCall(sqlStatement);
            return cstmt;
        }
        else
        {
            throw new DBCPException(ERR_POOL_MAXED_OUT);
        }
    }

    /**
     * Returns the Meta-Data of the underlying Connection object.
     */
    public DatabaseMetaData getMetaData() throws SQLException, DBCPException
    {
        if (conn != null)
        {
            return conn.getMetaData();
        }
        else
        {
            throw new DBCPException(ERR_POOL_MAXED_OUT);
        }
    }

    /**
     * Returns the underlying SQL Connection object. This connection must
     * <b>not</b> be closed directly.
     */
    public Connection getConnection()
    {
        return conn;
    }

    /**
     * Returns the connection back to the pool. Same as freeDBConnectionLib(),
     * except that this method does not throw an Exception.
     */
    public void freeDBConnection()
    {
        if (dbcp != null)
        {
            try
            {
                dbcp.freeDBConnection(dbConn);
            }
            catch (Exception ex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
                ex.printStackTrace();
            }
        }
    }

    public void freeDBConnectionLib() throws SQLException
    {
        if (dbcp != null)
        {
            dbcp.freeDBConnection(dbConn);
        }
    }

    /**
     * Returns the number of Connection objects currently available in the pool.
     */
    public int getAvailableCount()
    {
        int availableCount = 0;
        if (dbcp != null)
        {
            availableCount = dbcp.getAvailableCount();
        }
        return availableCount;
    }

}
