/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;
import net.infantai.nlp.ner.EntityTaggerOutput;
import net.infantai.util.common.TextUtil;
import net.infantai.util.common.TokenGroup;
import net.infantai.util.common.Tokenizer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class OrganizationNameTagger
{
    private static final Logger logger = LogManager.getLogger(OrganizationNameTagger.class);
    
    private static Set<String> organizationNameNearbyDescriptors = null;
    private static Set<String> organizationNameSuffixes = null;
    
    private static final Object objLock = new Object();
    
    private static final TextUtil tu = new TextUtil();
    
    private static final int MAX_TOKENS_IN_ORGANIZATION_NAME = 5;
    private static final double MAX_CONFIDENCE_SCORE = 0.85;
    private static final double PENALTY_FOR_NO_NEARBY_DESCRIPTOR_MATCH = 0.15;
    
    protected static final List<String> allDelim = new ArrayList<>();

    protected static final String TOKEN_DELIM_1 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".“”()[]{}`~|'‘’_‒–—\\/?=!+#*^&%©®™"; // splits on all common special characters
    protected static final String TOKEN_DELIM_2 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".[]()?"; // does not split on "&", "!"
    protected static final String TOKEN_DELIM_3 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞`~'\".[]/"; // does not split on "-", "(", ")"
    protected static final String TOKEN_DELIM_4 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞`~'\".[]/()"; // does not split on "-", but splits on "(", ")"
    protected static final String TOKEN_DELIM_5 = " <>	    \r\n:;,„`~'\"[]/-!()"; // does not split on ".", useful for tokens like "amazon.com"
    protected static final String TOKEN_DELIM_6 = " <>	    \r\n:;,„";    
    
    public OrganizationNameTagger()
    {
        synchronized(objLock)
        {
            if (organizationNameNearbyDescriptors == null)
            {
                organizationNameNearbyDescriptors = new LinkedHashSet<String>();
                organizationNameNearbyDescriptors.add("company");
                organizationNameNearbyDescriptors.add("firm");
                organizationNameNearbyDescriptors.add("maker");
                organizationNameNearbyDescriptors.add("manufacturer");
                organizationNameNearbyDescriptors.add("producer");
                organizationNameNearbyDescriptors.add("startup");
            }
            
            if (organizationNameSuffixes == null)
            {
                organizationNameSuffixes = new LinkedHashSet<String>();
                organizationNameSuffixes.add("ltd");
                organizationNameSuffixes.add("limited");
                organizationNameSuffixes.add("pvt ltd");
                organizationNameSuffixes.add("private limited");
                organizationNameSuffixes.add("plc");
                organizationNameSuffixes.add("pllc");
                organizationNameSuffixes.add("inc");
                organizationNameSuffixes.add("llc");
                organizationNameSuffixes.add("gmbh"); // Germany, Switzerland
                organizationNameSuffixes.add("ag"); // Germany, Switzerland
                organizationNameSuffixes.add("kg"); // Germany, Austria, Belgium etc.
                organizationNameSuffixes.add("gmbh co kg"); // Germany
                organizationNameSuffixes.add("kgaa"); // Germany
                organizationNameSuffixes.add("sa"); // france
                organizationNameSuffixes.add("sas"); // france
                organizationNameSuffixes.add("sarl");
                organizationNameSuffixes.add("sàrl");
                organizationNameSuffixes.add("corp");
            }            
            
            if (allDelim.isEmpty())
            {
                allDelim.add(TOKEN_DELIM_1);
                allDelim.add(TOKEN_DELIM_2);
                allDelim.add(TOKEN_DELIM_3);
                allDelim.add(TOKEN_DELIM_4);
                allDelim.add(TOKEN_DELIM_5);
                allDelim.add(TOKEN_DELIM_6);
            }            
        }
    }
    
    public List<EntityTaggerOutput> getOrganizationNames(String originalInput, List<TokenGroup> inputTokens)
    {
        List<EntityTaggerOutput> organizationNames = new ArrayList<EntityTaggerOutput>();
        Iterator<TokenGroup> iterTG = inputTokens.iterator();
        while (iterTG.hasNext())
        {
            TokenGroup tg = iterTG.next();
            if (organizationNameSuffixes.contains(tg.getToken().toLowerCase()))
            {
                
                String inputLeftOfTG = originalInput.substring(0, tg.getIndex());
                String[] prevTextParts = tu.getNTokensFromPrevText(inputLeftOfTG, MAX_TOKENS_IN_ORGANIZATION_NAME).split("\\W+");
                String lastQualifyingToken = "";
                String organizationNameParts = tg.getToken();
                for (int i = (prevTextParts.length - 1); i >= 0; --i)
                {
                    String token = prevTextParts[i];
                    if (token.length() > 0)
                    {
                        if (Character.isUpperCase(token.charAt(0)))
                        {
                            lastQualifyingToken = token;
                            organizationNameParts = token + " " + organizationNameParts;
                        }
                        else
                        {
                            break;
                        }                    
                    }
                }
                int startIndex = inputLeftOfTG.lastIndexOf(lastQualifyingToken);
                if (startIndex != -1)
                {
                    TokenGroup tgOrganizationName = new TokenGroup(organizationNameParts, startIndex, ((tg.getIndex() + tg.getSize()) - startIndex));
                    EntityTaggerOutput eto = new EntityTaggerOutput();
                    eto.tg = tgOrganizationName;
                    eto.isDictionaryMatch = false;
                    eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_ORGANIZATION;
                    eto.confidenceScore = MAX_CONFIDENCE_SCORE;
                    organizationNames.add(eto);
                    System.out.println("---- " + tgOrganizationName.getToken());
                }
            }
        }
        return organizationNames;
    }    
    
    public static void main(String[] args)
    {
        OrganizationNameTagger ont = new OrganizationNameTagger();
//        String input = "Tesla, Inc. (formerly Tesla Motors, Inc.) is an American electric vehicle and clean energy company based in Palo Alto, California. Tesla's current products include electric cars (the Model S, Model 3, Model X, and Model Y), battery energy storage from home to grid scale (the Powerwall, Powerpack, and Megapack), solar products (solar panels and solar roof tiles) and related products and services. Founded in July 2003 by engineers Martin Eberhard and Marc Tarpenning as Tesla Motors, the company’s name is a tribute to inventor and electrical engineer Nikola Tesla. The next three employees were Ian Wright, Elon Musk, and J. B. Straubel, all of whom are named as co-founders of the company. Elon Musk, who formerly served as chairman and is the current CEO, said in 2006 that \"the overarching purpose of Tesla Motors...is to help expedite the move from a mine-and-burn hydrocarbon economy towards a solar electric economy\" and it would build a wide range of electric vehicles, including \"affordably priced family cars\", and co-market SolarCity solar panels to do so. Tesla later acquired SolarCity in 2016.";
        String input = "ARNA Genomics US Inc, which develops tests for early stage cancer by detecting specific fragments of nucleic acids in plasma, announced today closing of the $3,5 million series A with a leading investor — Xploration Capital venture fund. Raised funding will be used for clinical validation of ARNA Breast — a first unique test for breast cancer.";
        try
        {
            List<TokenGroup> inputTGs = new Tokenizer().getAllTokens(input, allDelim.get(0), 6);
            ont.getOrganizationNames(input, inputTGs);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    
//    public static Set<TokenGroup> getAllDistinctTokens(String input, List<String> delimiters)
//    {
//        Set<TokenGroup> distinctTokens = new TreeSet<>();
//        Tokenizer tokenizer = new Tokenizer();
//        for (int j = 0; j < delimiters.size(); ++j)
//        {
//            distinctTokens.addAll(tokenizer.getAllTokens(input, delimiters.get(j), 6));
//        }
//        return distinctTokens;
//    }    
    
}
