/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import net.infantai.nlp.ner.EntityTaggerOutput;
import net.infantai.util.common.TextUtil;
import net.infantai.util.common.TokenBoundary;
import net.infantai.util.common.TokenGroup;
import opennlp.tools.namefind.NameFinderME;
import opennlp.tools.namefind.TokenNameFinderModel;
import opennlp.tools.tokenize.TokenizerME;
import opennlp.tools.tokenize.TokenizerModel;
import opennlp.tools.util.Span;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class PersonNameTagger
{
    private static final Logger logger = LogManager.getLogger(PersonNameTagger.class);
    
    private TokenNameFinderModel personNameModel = null;
    
    private TokenizerModel tokenizerModel = null;
    
    private static final TextUtil tu = new TextUtil();
    
    private static final int MAX_TOKENS_IN_PERSON_NAME = 3;
    private static final double MAX_CONFIDENCE_SCORE = 0.85;    
    
    private IWLLoader iwlLoader = IWLLoader.getInstance();
    
    private Set<String> firstNames = iwlLoader.getFirstNames();
    private Set<String> lastNames = iwlLoader.getLastNames();    
    private Set<String> allNames = new HashSet<>();
    
    protected static final String TOKEN_DELIM_1 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".“”()[]{}`~|'‘’_‒–—\\/?=!+#*^&%©®™"; // splits on all common special characters
    
    public PersonNameTagger()
    {
        InputStream personNameModelInStream = null;
        InputStream tokenizerInStream = null;
        
        allNames.addAll(firstNames);
        allNames.addAll(lastNames);
        
        try
        {        
            personNameModelInStream = new FileInputStream("./models/opennlp/en-ner-person.bin"); // from http://opennlp.sourceforge.net/models-1.5/
            personNameModel = new TokenNameFinderModel(personNameModelInStream);
            personNameModelInStream.close();

            tokenizerInStream = new FileInputStream("./models/opennlp/en-token.bin");
            tokenizerModel = new TokenizerModel(tokenizerInStream);            
            tokenizerInStream.close();
            
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
            fnfe.printStackTrace();
            logger.error(fnfe.getMessage(), fnfe);
        }
        catch (IOException ioex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
            ioex.printStackTrace();
            logger.error(ioex.getMessage(), ioex);
        }
        finally
        {
            try
            {
                if (personNameModelInStream != null)
                {
                    personNameModelInStream.close();
                }
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                logger.error(ioex.getMessage(), ioex);                
            }
            
            try
            {
                if (tokenizerInStream != null)
                {
                    tokenizerInStream.close();
                }
            }
            catch (IOException ioex)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                ioex.printStackTrace();
                logger.error(ioex.getMessage(), ioex);                
            }            
        }
    }
    
    public List<EntityTaggerOutput> getPersonNames(String originalInput, Set<TokenGroup> inputTokens)
    {
        List<EntityTaggerOutput> personNames = new ArrayList<EntityTaggerOutput>();
        Iterator<TokenGroup> iterTG = inputTokens.iterator();
        Map<Integer, EntityTaggerOutput> indexesProcessed = new LinkedHashMap<>();
        List<TokenBoundary> tokenBoundaries = new ArrayList<TokenBoundary>();
        while (iterTG.hasNext())
        {
            TokenGroup tg = iterTG.next();
            if (allNames.contains(tg.getToken().toLowerCase()))
            {
                
                String inputLeftOfTG = originalInput.substring(0, tg.getIndex());
                String[] prevTextParts = tu.getNTokensFromPrevText(inputLeftOfTG, MAX_TOKENS_IN_PERSON_NAME - 1).split("\\W+");
                String lastQualifyingToken = "";
                String personNameParts = tg.getToken();
                for (int i = (prevTextParts.length - 1); i >= 0; --i)
                {
                    String token = prevTextParts[i];
                    if (token.length() > 0)
                    {
                        String tokenLCase = token.toLowerCase();
                        if (Character.isUpperCase(token.charAt(0)))
                        {
                            if (allNames.contains(tokenLCase) || token.length() == 1)
                            {
                                lastQualifyingToken = token;
                                personNameParts = token + " " + personNameParts;                        
                            }
                        }
                        else
                        {
                            break;
                        }                    
                    }
                }
                int startIndex = inputLeftOfTG.lastIndexOf(lastQualifyingToken);
                if (startIndex != -1)
                {
                    if (originalInput.indexOf(personNameParts) == startIndex)
                    {
                        EntityTaggerOutput etoSameIndex = indexesProcessed.get(new Integer(startIndex));
                        TokenBoundary tbound = new TokenBoundary();
                        tbound.startIndex = startIndex;
                        tbound.endIndex = (tg.getIndex() + tg.getSize());                        
                        TokenGroup tgPersonName = new TokenGroup(personNameParts, startIndex, ((tg.getIndex() + tg.getSize()) - startIndex));
                        EntityTaggerOutput eto = new EntityTaggerOutput();
                        eto.tg = tgPersonName;
                        eto.isDictionaryMatch = false;
                        eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_PERSON;
                        eto.confidenceScore = MAX_CONFIDENCE_SCORE;                        
                        if (etoSameIndex == null)
                        {
                            // new start index
                            if (!TextUtil.hasOverlap(tokenBoundaries, startIndex, tbound.endIndex - startIndex))
                            {
                                indexesProcessed.put(startIndex, eto);
                                System.out.println("---- " + tgPersonName.getToken());                    
                                tokenBoundaries.add(tbound);                                
                            }
                            else
                            {
                                System.out.println("Skipping " + tgPersonName + " due to overlap.");
                            }
                        }
                        else 
                        {
                            if (etoSameIndex.tg.getToken().length() < personNameParts.length())
                            {
//                                TokenGroup tgPersonName = new TokenGroup(personNameParts, startIndex, ((tg.getIndex() + tg.getSize()) - startIndex));
//                                EntityTaggerOutput eto = new EntityTaggerOutput();
//                                eto.tg = tgPersonName;
//                                eto.isDictionaryMatch = false;
//                                eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_PERSON;
//                                eto.confidenceScore = MAX_CONFIDENCE_SCORE;
                                indexesProcessed.put(startIndex, eto);
                                System.out.println("---- " + tgPersonName.getToken());                    
                                tokenBoundaries.add(tbound);
                            }
                        }
                    }
                }
            }
        }
        for (Map.Entry<Integer, EntityTaggerOutput> mapEntry : indexesProcessed.entrySet())
        {
            if (!TextUtil.hasOverlap(tokenBoundaries, mapEntry.getValue().tg.getIndex(), mapEntry.getValue().tg.getSize()))
            {
                personNames.add(mapEntry.getValue());
            }
            else
            {
                System.out.println("Skipping " + mapEntry.getValue().tg.getToken() + " due to overlap.");
            }            
        }
        return personNames;
    }        
    
    public List<EntityTaggerOutput> getPersonNamesOpenNLP(String input)
    {
        List<EntityTaggerOutput> personNames = new ArrayList<EntityTaggerOutput>();
        TokenizerME tokenizer = new TokenizerME(tokenizerModel);
        NameFinderME nameFinder = new NameFinderME(personNameModel);
//        String[] tokens = paragraph.split("\\W+");
        String[] tokens = tokenizer.tokenize(input);
        Span[] nameSpans = nameFinder.find(tokens);
        String[] names = Span.spansToStrings(nameSpans, tokens);
        if (nameSpans.length == names.length)
        {
            int lastIndex = 0;
            for (int i = 0; i < nameSpans.length; ++i)
            {
                Span nameSpan = nameSpans[i];
                EntityTaggerOutput eto = new EntityTaggerOutput();
                if (lastIndex < input.length())
                {
                    int startIndex = input.indexOf(names[i], lastIndex); // not sure why, but OpenNLP Span getStart() returns an incorrect index
                    if (startIndex != -1)
                    {
                        lastIndex = startIndex + 1;
                        System.out.println(names[i] + " start: " + nameSpans[i].getStart() + " end: " + nameSpans[i].getEnd() + " computed start: " + startIndex);
                        TokenGroup tgName = new TokenGroup(names[i], startIndex, names[i].length());
                        eto.tg = tgName;
                        eto.isDictionaryMatch = false;
                        eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_PERSON;
                        eto.confidenceScore = nameSpan.getProb();
                        personNames.add(eto);                                    
                    }
                }
            }        
        }
        return personNames;
    }    
    
    public static void main(String[] args)
    {
        PersonNameTagger pnt = new PersonNameTagger();
        String input = "Ashish Nawal Saharia of Tesla, Inc. (formerly Tesla Motors, Inc.) is an American electric vehicle and clean energy company based in Palo Alto, California. Tesla's current products include electric cars (the Model S, Model 3, Model X, and Model Y), battery energy storage from home to grid scale (the Powerwall, Powerpack, and Megapack), solar products (solar panels and solar roof tiles) and related products and services. Founded in July 2003 by engineers Martin Eberhard and Marc Tarpenning as Tesla Motors, the company’s name is a tribute to inventor and electrical engineer Nikola Tesla. The next three employees were Ian Wright, Elon Musk, and J. B. Straubel, all of whom are named as co-founders of the company. Elon Musk, who formerly served as chairman and is the current CEO, said in 2006 that \"the overarching purpose of Tesla Motors...is to help expedite the move from a mine-and-burn hydrocarbon economy towards a solar electric economy\" and it would build a wide range of electric vehicles, including \"affordably priced family cars\", and co-market SolarCity solar panels to do so. Tesla later acquired SolarCity in 2016.";
        try
        {
//            List<EntityTaggerOutput> lst = pnt.getPersonNamesOpenNLP(input);
//            for (EntityTaggerOutput eto : lst)
//            {
//                System.out.println("Name: " + eto.tg.getToken() + " Type: " + eto.entityType + " Confidence: " + eto.confidenceScore);
//            }
            
            List<String> allDelim = new ArrayList<>();
            allDelim.add(TOKEN_DELIM_1);
            Set<TokenGroup> inputTGs = TextUtil.getAllDistinctTokens(input, allDelim, 4);
            List<EntityTaggerOutput> lst2 = pnt.getPersonNames(input, inputTGs);
            for (EntityTaggerOutput eto : lst2)
            {
                System.out.println("Name: " + eto.tg.getToken() + " Type: " + eto.entityType + " Confidence: " + eto.confidenceScore);
            }
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    
}
