/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.nlp.ner;

import java.util.ArrayList;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import net.infantai.nlp.NegationDetector;
import net.infantai.nlp.OrganizationNameTagger;
import net.infantai.nlp.POSTagger;
import net.infantai.nlp.PersonNameTagger;
import net.infantai.nlp.WordNetUtil;
import net.infantai.util.common.BooleanTextRuleMatcher;
import net.infantai.util.common.CommonUtil;
import net.infantai.util.common.DateFinder;
import net.infantai.util.common.DateParts;
import net.infantai.util.common.MalformedExpressionException;
import net.infantai.util.common.TextUtil;
import net.infantai.util.common.TokenBoundary;
import net.infantai.util.common.TokenGroup;
import net.infantai.util.common.Tokenizer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class EntityTagger
{

    private static final Logger LOGGER = LogManager.getLogger(EntityTagger.class);

    protected static final List<String> allDelim = new ArrayList<>();

    protected static final String TOKEN_DELIM_1 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".“”()[]{}`~|'‘’_‒–—\\/?=!+#*^&%©®™"; // splits on all common special characters
    protected static final String TOKEN_DELIM_2 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞-\".[]()?"; // does not split on "&", "!"
    protected static final String TOKEN_DELIM_3 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞`~'\".[]/"; // does not split on "-", "(", ")"
    protected static final String TOKEN_DELIM_4 = " <>	    \r\n:;,‚„ʻʽ‛‟❛❜❝❞`~'\".[]/()"; // does not split on "-", but splits on "(", ")"
    protected static final String TOKEN_DELIM_5 = " <>	    \r\n:;,„`~'\"[]/-!()"; // does not split on ".", useful for tokens like "amazon.com"
    protected static final String TOKEN_DELIM_6 = " <>	    \r\n:;,„";

    protected static int MAX_TOKEN_COUNT = 6;
    protected static int MAX_TOKEN_GROUP_LENGTH = 4;
    protected static final boolean ALLOW_PREPOSITIONS = true;
    protected static final boolean ALLOW_CONJUNCTIONS = true;
    protected static final boolean ALLOW_ADJECTIVES = true;
    protected static final boolean ALLOW_DETERMINERS = true;
    protected static final boolean ALLOW_ADVERBS = false;

    protected static int MIN_DATE_YEAR = 1981;
    protected static int MAX_DATE_YEAR = 3001; // Space Odyssey !

//    private static long ENTITY_ID_FOR_DATE = -2L;
//    private static long ENTITY_TYPE_BROKER = 56L;
//    private static long ENTITY_TYPE_COMPANY = 1L;

    private static final POSTagger postagger = new POSTagger();

    private static final Set<String> digitsAndUpperCaseLetters = new LinkedHashSet<>();

//    private static SentimentTagger sentimentTagger = null;
    private DateFinder dateFinder = null;
    private WordNetUtil wnutil = null;
    private TextUtil textUtil = null;
    private PersonNameTagger personNameTagger = new PersonNameTagger();
    private OrganizationNameTagger organizationNameTagger = new OrganizationNameTagger();

//    private static Set<Long> entityTypeIdsToIgnoreForSentimentTagging = new LinkedHashSet<>();
    
//    private static Set<String> brokerVerbsToCheck = new LinkedHashSet<String>();

    private static final Object constructorLock = new Object();
    
    private Set<String> negationPreMarkers = new LinkedHashSet<>(); // default negation markers
    private int maxNegationTokensPre = 6; 
    private Set<String> negationPostMarkers = new LinkedHashSet<>();  // default negation markers
    private int maxNegationTokensPost = 6;
    
    private boolean enableExperimentalPlaceholderTagging = true;
    private boolean trimPlaceholderValueInKeyword = true;
    private Set<String> tokensToTrimInPlaceholderKeywords = new LinkedHashSet<>();
    private Pattern patNumericValue = null;

    public EntityTagger()
    {
        synchronized (constructorLock)
        {
            if (allDelim.isEmpty())
            {
                allDelim.add(TOKEN_DELIM_1);
                allDelim.add(TOKEN_DELIM_2);
                allDelim.add(TOKEN_DELIM_3);
                allDelim.add(TOKEN_DELIM_4);
                allDelim.add(TOKEN_DELIM_5);
                allDelim.add(TOKEN_DELIM_6);
            }

            if (digitsAndUpperCaseLetters.isEmpty())
            {
                digitsAndUpperCaseLetters.add("0");
                digitsAndUpperCaseLetters.add("1");
                digitsAndUpperCaseLetters.add("2");
                digitsAndUpperCaseLetters.add("3");
                digitsAndUpperCaseLetters.add("4");
                digitsAndUpperCaseLetters.add("5");
                digitsAndUpperCaseLetters.add("6");
                digitsAndUpperCaseLetters.add("7");
                digitsAndUpperCaseLetters.add("8");
                digitsAndUpperCaseLetters.add("9");

                digitsAndUpperCaseLetters.add("A");
                digitsAndUpperCaseLetters.add("B");
                digitsAndUpperCaseLetters.add("C");
                digitsAndUpperCaseLetters.add("D");
                digitsAndUpperCaseLetters.add("E");
                digitsAndUpperCaseLetters.add("F");
                digitsAndUpperCaseLetters.add("G");
                digitsAndUpperCaseLetters.add("H");
                digitsAndUpperCaseLetters.add("I");
                digitsAndUpperCaseLetters.add("J");
                digitsAndUpperCaseLetters.add("K");
                digitsAndUpperCaseLetters.add("L");
                digitsAndUpperCaseLetters.add("M");
                digitsAndUpperCaseLetters.add("N");
                digitsAndUpperCaseLetters.add("O");
                digitsAndUpperCaseLetters.add("P");
                digitsAndUpperCaseLetters.add("Q");
                digitsAndUpperCaseLetters.add("R");
                digitsAndUpperCaseLetters.add("S");
                digitsAndUpperCaseLetters.add("T");
                digitsAndUpperCaseLetters.add("U");
                digitsAndUpperCaseLetters.add("V");
                digitsAndUpperCaseLetters.add("W");
                digitsAndUpperCaseLetters.add("X");
                digitsAndUpperCaseLetters.add("Y");
                digitsAndUpperCaseLetters.add("Z");
            }
//            if (brokerVerbsToCheck.isEmpty())
//            {
//                brokerVerbsToCheck.add("says");
//                brokerVerbsToCheck.add("say");
//            }
//            if(entityTypeIdsToIgnoreForSentimentTagging.isEmpty())
//            {
//                entityTypeIdsToIgnoreForSentimentTagging.add(ENTITY_TYPE_BROKER);
//            }

            if (negationPreMarkers.isEmpty())
            {
                negationPreMarkers.add("no");
                negationPreMarkers.add("no visible");
                negationPreMarkers.add("absence of");
                negationPreMarkers.add("not");
                negationPreMarkers.add("not have");
                negationPreMarkers.add("not have any");
                negationPreMarkers.add("not having");
                negationPreMarkers.add("not had");
                negationPreMarkers.add("not had any");
                negationPreMarkers.add("never had");
                negationPreMarkers.add("never had any");
                negationPreMarkers.add("no sign of");
                negationPreMarkers.add("hasn't had");
                negationPreMarkers.add("hasn't had any");
                negationPreMarkers.add("no history of");
                negationPreMarkers.add("no complaint of");
                negationPreMarkers.add("never complained of");
                negationPreMarkers.add("does not complaint of");
                negationPreMarkers.add("never had any history of");
                negationPreMarkers.add("never experienced");
                negationPreMarkers.add("not on");
                negationPreMarkers.add("not taking");
                negationPreMarkers.add("not take");
                negationPreMarkers.add("not taken");
                negationPreMarkers.add("showing no sign of");
                negationPreMarkers.add("showing no signs of");
                negationPreMarkers.add("showing no symptom of");
                negationPreMarkers.add("showing no symptoms of");
                negationPreMarkers.add("showing no visible sign of");
                negationPreMarkers.add("showing no visible signs of");
                negationPreMarkers.add("not showing any");
                negationPreMarkers.add("not displaying any");
                negationPreMarkers.add("not experiencing any");
                negationPreMarkers.add("not showing any sign of");
                negationPreMarkers.add("not showing any signs of");
                negationPreMarkers.add("not showing any symptom of");
                negationPreMarkers.add("not showing any symptoms of");
                negationPreMarkers.add("not displaying any sign of");
                negationPreMarkers.add("not displaying any signs of");
                negationPreMarkers.add("not displaying any symptom of");
                negationPreMarkers.add("not displaying any symptoms of");   
                negationPreMarkers.add("not experiencing any sign of");
                negationPreMarkers.add("not experiencing any signs of");
                negationPreMarkers.add("not experiencing any symptom of");
                negationPreMarkers.add("not experiencing any symptoms of");   
                negationPreMarkers.add("unable to");   
                negationPreMarkers.add("without");
                negationPreMarkers.add("w/o");
                negationPreMarkers.add("never been admitted for");
                negationPreMarkers.add("never been admitted for any");
                negationPreMarkers.add("never been on");
                negationPreMarkers.add("not been on");
                negationPreMarkers.add("never been on any");
                negationPreMarkers.add("not been on any");
            }
            
            if (negationPostMarkers.isEmpty())
            {
                negationPostMarkers.add("not observed");
                negationPostMarkers.add("not seen");
                negationPostMarkers.add("absent");
                negationPostMarkers.add("is absent");
                negationPostMarkers.add("is not observed");
                negationPostMarkers.add("was not observed");
                negationPostMarkers.add("isn't observed");
                negationPostMarkers.add("are not observed");
                negationPostMarkers.add("is not seen");
                negationPostMarkers.add("was not seen");
                negationPostMarkers.add("isn't seen");
                negationPostMarkers.add("are not seen");
                negationPostMarkers.add("is not found");
                negationPostMarkers.add("was not found");
                negationPostMarkers.add("isn't found");
                negationPostMarkers.add("are not found");                
                negationPostMarkers.add("not present");
                negationPostMarkers.add("is not present");
                negationPostMarkers.add("was not present");
                negationPostMarkers.add("isn't present");
                negationPostMarkers.add("are not present");
                negationPostMarkers.add("not found");
                negationPostMarkers.add("not visible");
                negationPostMarkers.add("is not found");
                negationPostMarkers.add("was not found");
                negationPostMarkers.add("isn't found");
                negationPostMarkers.add("are not found");
                negationPostMarkers.add("not located");
                negationPostMarkers.add("is not located");
                negationPostMarkers.add("was not located");
                negationPostMarkers.add("isn't located");
                negationPostMarkers.add("are not located");
                negationPostMarkers.add("could not be observed");
                negationPostMarkers.add("could not be found");
                negationPostMarkers.add("could not be located");
                
                negationPostMarkers.add("ruled out");
                negationPostMarkers.add("is ruled out");
                negationPostMarkers.add("was ruled out");
                negationPostMarkers.add("were ruled out");
            }
            
            if (patNumericValue == null)
            {
                patNumericValue = Pattern.compile("-?[0123456789,]+\\.?[0123456789]*");
            }
            
            if (tokensToTrimInPlaceholderKeywords.isEmpty())
            {
                tokensToTrimInPlaceholderKeywords.add("is");
                tokensToTrimInPlaceholderKeywords.add("was");
            }

        }
        
        textUtil = new TextUtil();
//        sentimentTagger = SentimentTagger.getInstance();
        dateFinder = DateFinder.getInstance();
        wnutil = WordNetUtil.getInstance();
    }

    public static TokenGroup[] getTokensForOpenNLPPOSTagging(String input)
    {
        Tokenizer tokenizer = new Tokenizer();
        List<TokenGroup> tokens = tokenizer.getAllTokens(input, " <>	    \r\n;:,‚„ʻʽ‛‟❛❜❝❞./-+?!`~|'\"“”‘’_‒–—{}[]()\\&*#*©®™", 1);
        TokenGroup[] tgarr = new TokenGroup[tokens.size()];
        return tokens.toArray(tgarr);
    }

    public List<EntityTaggerOutput> tagEntities(String input, Long inputId, boolean matchCaseSensitiveKeywords, boolean tagNonDictionaryEntities, boolean usePOSTagFiltering, boolean tagDates, Map<String, Set<EntityKeyword>> entityKeywordsCaseInsensitive, Map<String, Set<EntityKeyword>> entityKeywordsCaseSensitive, Set<String> unlikelyEntityNames)
    {
        long t1, t2;
        t1 = System.nanoTime();
//        StaticLoader sloader = StaticLoader.getInstance();
        Map<TokenGroup, List<EntityKeyword>> matches = new LinkedHashMap<>();
        Set<String> allMatchedStrings = new LinkedHashSet<>();
        List<TokenBoundary> boundaries = new ArrayList<>();
        String inputLCase = input.toLowerCase();
//        Set<Long> brokerCompanyEntityIds = sloader.getBrokerCompanyEntityIds();
//        Map<String, Set<EntityKeyword>> entityKeywordsCaseInsensitive = sloader.getEntityKeywordsCaseInsensitive();
//        Map<String, Set<EntityKeyword>> entityKeywordsCaseSensitive = sloader.getEntityKeywordsCaseSensitive();
//        Set<String> unlikelyEntityNames = sloader.getUnlikelyEntityNames();
//        Set<String> brokerIdentificationPreTokens = sloader.getBrokerIdentificationPreTokens();
//        Set<String> brokerIdentificationPostTokens = sloader.getBrokerIdentificationPostTokens();
        List<String> unknownEntities = new ArrayList<>(); // only for logging
        Map<TokenGroup, String> nouns = null;
        List<EntityTaggerOutput> etoList = new ArrayList<>();
//        List<EntityTaggerOutput> esopList = new ArrayList<>();
//        List<EntityTaggerOutput> dateEsopList = new ArrayList<>();
//        boolean isBrokerTagged = false;
        Set<TokenGroup> inputTokenGroups = TextUtil.getAllDistinctTokens(input, allDelim, MAX_TOKEN_COUNT); // retain character case of input - do not convert to lower / upper case here
        Set<String> inputTokensForBTRM = getAllTokensForBTRM(inputTokenGroups); // implicit lower case conversion of tokens        
        Iterator<TokenGroup> inputTokenGroupsIter = inputTokenGroups.iterator();
        while (inputTokenGroupsIter.hasNext())
        {
            TokenGroup tg = inputTokenGroupsIter.next();
            String token = tg.getToken();
            String tokenLCase = token.toLowerCase();
            Integer tindex = tg.getIndex();

            int tokenLengthAdjustment = 0;

            boolean skip = false;

            Iterator<Map.Entry<TokenGroup, List<EntityKeyword>>> iterPreviousMatches = matches.entrySet().iterator();
            while (iterPreviousMatches.hasNext())
            {
                Map.Entry<TokenGroup, List<EntityKeyword>> mapEntry = iterPreviousMatches.next();
                TokenGroup tgAlreadyMatched = mapEntry.getKey();

                // Check if this match is contained within a previous match. This is important to 
                // ensure that "Bank of India" doesn't match with the same part of the string that yielded a 
                // successful match for "State Bank of India".
//                if (tindex.intValue() >= tgAlreadyMatched.getIndex().intValue() && tindex.intValue() <= (tgAlreadyMatched.getIndex().intValue() + tgAlreadyMatched.getSize().intValue()))
                if (tindex.intValue() >= tgAlreadyMatched.getIndex().intValue() && tindex.intValue() <= (tgAlreadyMatched.getIndex().intValue() + tgAlreadyMatched.getSize().intValue()) && (tg.getIndex().intValue() + tg.getSize().intValue()) <= (tgAlreadyMatched.getIndex().intValue() + tgAlreadyMatched.getSize().intValue()))
                {
                    // Note that only those matches which are entirely contained within previous matches will be discarded at this stage. 
                    // Matches which start inside but end outside of a previous match will be allowed, to be potentially handled by the OverlapResolver later.
                    skip = true;
                }
            }

            if (skip)
            {
                LOGGER.debug("Skipping possible match, overlapping previous match found: " + token + ", index: " + tindex);
                continue;
            }

            Set<EntityKeyword> entityKeywords = entityKeywordsCaseInsensitive.get(tokenLCase);
            if (entityKeywords == null)
            {
                entityKeywords = new LinkedHashSet<>();
            }

            if (matchCaseSensitiveKeywords)
            {
                Set<EntityKeyword> entityKeywordsCS = entityKeywordsCaseSensitive.get(token);
                if (entityKeywordsCS != null)
                {
                    entityKeywords.addAll(entityKeywordsCS);
                }
            }
            
            // experimental code to process infant placeholders
            if (enableExperimentalPlaceholderTagging)
            {
                String placeholderSubstitutedToken = tokenLCase.replaceAll("-?[0123456789,]+\\.?[0123456789]*", "{infant-placeholder#number}");
                Set<EntityKeyword> entityKeywordsPlaceholder = entityKeywordsCaseInsensitive.get(placeholderSubstitutedToken);
                if (entityKeywordsPlaceholder != null)
                {
                    Set<EntityKeyword> ekPlaceholderSet = new LinkedHashSet<>();
                    for (EntityKeyword ek : entityKeywordsPlaceholder)
                    {
                        EntityKeyword ekSubstituted = new EntityKeyword(ek);
                        ekSubstituted.setKeyword(token);
                        ekSubstituted.setIsModification(true);
                        ekSubstituted.setAttemptNumericValueExtraction(true);
                        ekPlaceholderSet.add(ekSubstituted);
                    }
                    entityKeywords.addAll(ekPlaceholderSet);
                }
            }
            //

            if (!entityKeywords.isEmpty())
            {
                boolean foundAtLeastOneMatch = false;
                Iterator<EntityKeyword> iterEntityKeywords = entityKeywords.iterator();
//                for (int i = 0; i < entityKeywords.size(); ++i)
                while (iterEntityKeywords.hasNext())
                {
                    boolean matchFound = true;
                    EntityKeyword entityKeyword = iterEntityKeywords.next();
                    BooleanTextRuleMatcher btrm = entityKeyword.getKeywordRule();
                    if (btrm != null)
                    {
                        matchFound = false;
                        try
                        {
                            matchFound = btrm.matchBooleanTextRule(inputTokensForBTRM);
                        }
                        catch (MalformedExpressionException mex)
                        {
                            LOGGER.error("MalformedExpressionException => keyword id: " + entityKeyword.getKeywordId() + " Rule: " + btrm.toString(), mex);
                        }
                        catch (Exception ex)
                        {
                            LOGGER.error(ex.getMessage() + " => keyword id: " + entityKeyword.getKeywordId() + " Rule: " + btrm.toString(), ex);
                        }
                    }
                    if (matchFound)
                    {
                        foundAtLeastOneMatch = true;
                        List<EntityKeyword> taggedEntityKeywords = matches.get(tg);
                        if (taggedEntityKeywords == null)
                        {
                            taggedEntityKeywords = new ArrayList<>();
                            matches.put(tg, taggedEntityKeywords);
                        }
                        taggedEntityKeywords.add(entityKeyword);

                        // add to collection etoList
                        EntityTaggerOutput eto = new EntityTaggerOutput();
                        eto.ek = entityKeyword;
                        eto.tg = tg;
                        eto.isDictionaryMatch = true;
                        
                        TokenGroup tgNegation = NegationDetector.getNegationText(inputLCase, eto, negationPreMarkers, maxNegationTokensPre, negationPostMarkers, maxNegationTokensPost);
                        eto.tgNegation = tgNegation;
                        
                        // experimental code to attempt numeric value extraction from keyword
                        if (entityKeyword.attemptNumericValueExtraction())
                        {
                            Matcher mat = patNumericValue.matcher(entityKeyword.getKeyword());
                            if (mat.find()) // no need to loop here since multiple placeholders are not expected in a single keyword
                            {
                                eto.possibleValue = mat.group(0);
                                int possibleValueIndex = entityKeyword.getKeyword().indexOf(eto.possibleValue);
                                if (possibleValueIndex != -1)
                                {
                                    
                                    if (trimPlaceholderValueInKeyword)
                                    {
                                        EntityKeyword ek = eto.ek;
                                        LOGGER.debug("Discarding possible value from keyword: " + ek.getKeyword() + ", index (within keyword): " + possibleValueIndex);
                                        String keywordWithoutValue = "";
                                        if (possibleValueIndex > 0)
                                        {
                                            keywordWithoutValue = ek.getKeyword().substring(0, possibleValueIndex).trim();
                                            keywordWithoutValue = textUtil.trimTokens(keywordWithoutValue, tokensToTrimInPlaceholderKeywords);
                                            
                                            if (keywordWithoutValue.length() > 0)
                                            {
                                                ek.setKeyword(keywordWithoutValue);
                                                // also modify tg
                                                eto.tg.setToken(ek.getKeyword());
                                                eto.tg.setSize(ek.getKeyword().length());
                                            }
                                        }
                                        else
                                        {
                                            // keyword begins with value
                                            if (eto.possibleValue.length() < ek.getKeyword().length())
                                            {
                                                keywordWithoutValue = ek.getKeyword().substring(possibleValueIndex).trim();
                                                keywordWithoutValue = textUtil.trimTokens(keywordWithoutValue, tokensToTrimInPlaceholderKeywords);
                                                
                                                if (keywordWithoutValue.length() > 0)
                                                {
                                                    int modifiedTGIndex = ek.getKeyword().indexOf(keywordWithoutValue, eto.possibleValue.length());
                                                    if (modifiedTGIndex > 0)
                                                    {
                                                        LOGGER.debug("Original TG: " + eto.tg);
                                                        ek.setKeyword(keywordWithoutValue);
                                                        // also modify tg
                                                        modifiedTGIndex = eto.tg.getIndex() + modifiedTGIndex;
                                                        eto.tg.setToken(ek.getKeyword());
                                                        eto.tg.setSize(ek.getKeyword().length());
                                                        eto.tg.setIndex(modifiedTGIndex);
                                                        LOGGER.debug("Modified TG: " + eto.tg);
                                                    }                                                
                                                }
                                            }
                                        }
                                    }
                                    
                                    possibleValueIndex = eto.tg.getIndex() + possibleValueIndex;
                                    eto.possibleValueIndex = possibleValueIndex;
                                    
                                    LOGGER.debug("Found possible value in placeholder keyword [" + eto.ek.getKeyword()+ "] => value: " + eto.possibleValue + ", index: " + eto.possibleValueIndex + ", original keyword: " + eto.ek.getKeywordOriginal());
                                }
                                else
                                {
                                    // highly unlikely that the program flow will ever get here
                                    LOGGER.debug("Discarding placeholder extracted value as index could not be determined: " + eto.possibleValue);
                                    eto.possibleValue = null;
                                }
                            }
                        }
                        //
                        
                        etoList.add(eto);
//                        if (entityKeyword.getEntityTypeId() == ENTITY_TYPE_BROKER)
//                        {
//                            isBrokerTagged = true;
//                        }
                    }
                }

                if (foundAtLeastOneMatch)
                {
//                    allMatchedStrings.add(tg.getToken()); // To avoid duplication through POS tagging. Edit (20181107 - We actually need the duplicates too; "Jetta" should be allowed as an Unknown Entity, even if it exists as a keyword for Volkswagen.)
                    TokenBoundary tokenBoundary = new TokenBoundary();
                    tokenBoundary.startIndex = tg.getIndex();
                    tokenBoundary.endIndex = tg.getIndex() + tg.getSize();
//                    boundaries.add(tokenBoundary);  // Commenting this out to allow Unknown Entities to be formed even within Dictionary Entity boundaries.
                }

//                matches.put(tg, entityKeywords);
//                allMatchedStrings.add(tg.getToken()); // to avoid duplication through POS tagging
//                TokenBoundary tokenBoundary = new TokenBoundary();
//                tokenBoundary.startIndex = tg.getIndex();
//                tokenBoundary.endIndex = tg.getIndex() + tg.getSize();
//                boundaries.add(tokenBoundary);
                // add to collection for sentiment analysis
//                for (int i = 0; i < entityKeywords.size(); ++i)
//                {
//                    EntityTaggerOutput eto = new EntityTaggerOutput();
//                    eto.ek = entityKeywords.get(i);
//                    eto.tg = tg;
//                    eto.isDictionaryMatch = true;
//                    etoList.add(eto);
//                }
            }
            else 
            {
                /**
                 * This block is not needed presently. Can be used to detect non-dictionary entities later.
                if (tagNonDictionaryEntities && !usePOSTagFiltering)
                {
                    unknownEntities.add(tg.getToken().toLowerCase());
                    EntityTaggerOutput eto = new EntityTaggerOutput();
                    //                            eto.ek = new EntityKeyword(); // object created with new EntityKeyword instance.
                    eto.tg = tg;
                    eto.isDictionaryMatch = false;
                    etoList.add(eto);
                }
                */
            }
        }

        if (tagDates)
        {
            Map<Integer, String> allDates = dateFinder.getDateStrings(input);
            Iterator<Map.Entry<Integer, String>> allDatesIter = allDates.entrySet().iterator();
            while (allDatesIter.hasNext())
            {
                boolean skip = false;
                Map.Entry<Integer, String> mapentry = allDatesIter.next();
                Integer index = mapentry.getKey();
                
//                Iterator<Map.Entry<TokenGroup, List<EntityKeyword>>> iterPreviousMatches = matches.entrySet().iterator();
//                while (iterPreviousMatches.hasNext())
//                {
//                    Map.Entry<TokenGroup, List<EntityKeyword>> mapEntry = iterPreviousMatches.next();
//                    TokenGroup tgAlreadyMatched = mapEntry.getKey();
//
//                    // Check if this match is contained within a previous match. This is important to 
//                    // ensure that "Bank of India" doesn't match with the same part of the string that yielded a 
//                    // successful match for "State Bank of India".
//                    if (index.intValue() >= tgAlreadyMatched.getIndex().intValue() && index.intValue() <= (tgAlreadyMatched.getIndex().intValue() + tgAlreadyMatched.getSize().intValue()))
//                    {
//                        skip = true;
//                    }
//                }
//
//                if (skip)
//                {
//                    continue;
//                }                
                
                
                String matchedString = mapentry.getValue();
                DateParts dateparts = dateFinder.interpretDateString(-1, matchedString);
                if (dateparts != null)
                {
                    if (dateparts.getYYYY() >= MIN_DATE_YEAR && dateparts.getYYYY() <= MAX_DATE_YEAR)
                    {
                        if (dateparts.getDD() > 0 && dateparts.getMM() > 0 && dateparts.getYYYY() > 0)
                        {
                            GregorianCalendar cal = new GregorianCalendar(dateparts.getYYYY(), dateparts.getMM() - 1, dateparts.getDD());
                            dateparts.setDate(cal);
                        }
//                        EntityTaggerOutput dateEsop = new EntitySentimentOutput();
//                        dateEsop.setEntityId(ENTITY_ID_FOR_DATE);
//                        dateEsop.setEntityKeyword(matchedString);
//                        dateEsop.setEntityKeywordIndex(index);
//                        dateEsop.setEntityDateParts(dateparts);
//                        dateEsopList.add(dateEsop);

                        EntityTaggerOutput etoDate = new EntityTaggerOutput();
                        //                            eto.ek = new EntityKeyword(); // object created with new EntityKeyword instance.
                        String matchedStringWithoutTime = dateFinder.discardTimePartFromDatetimeString(matchedString); // experimental / optional
                        TokenGroup tg = new TokenGroup(matchedStringWithoutTime, (Integer) index, (Integer) matchedStringWithoutTime.length());
                        
                        if (!TextUtil.hasOverlap(boundaries, tg.getIndex(), tg.getSize()))
                        {
                            etoDate.tg = tg;
                            etoDate.entityDateParts = dateparts;
                            etoDate.isDictionaryMatch = false;
                            etoDate.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_DATE;
                            
                            TokenGroup tgNegation = NegationDetector.getNegationText(inputLCase, etoDate, negationPreMarkers, maxNegationTokensPre, negationPostMarkers, maxNegationTokensPost);
                            etoDate.tgNegation = tgNegation;                            
                            
                            etoList.add(etoDate);

    //                        EntityKeyword pseudoEK = new EntityKeyword();
    //                        pseudoEK.setEntityName("PSEUDO_DATE");
    //                        pseudoEK.setKeyword(matchedString);
    //                        List<EntityKeyword> pseudoEKList = new ArrayList<>();
    //                        pseudoEKList.add(pseudoEK);
    //                        matches.put(tg, pseudoEKList);

                            allMatchedStrings.add(matchedString);
                            TokenBoundary tokenBoundary = new TokenBoundary();
                            tokenBoundary.startIndex = index;
                            tokenBoundary.endIndex = index + matchedString.length();
                            boundaries.add(tokenBoundary);                        
                        }
                        
                    }
                }
            }

        }

        // POS tagging
        if (tagNonDictionaryEntities)
        {
            
            List<EntityTaggerOutput> personNames = personNameTagger.getPersonNamesOpenNLP(input);
            Map<String, Double> personNamesConfidence = new LinkedHashMap<>();
            for (int i = 0; i < personNames.size(); ++i)
            {
                personNamesConfidence.put(personNames.get(i).tg.getToken().toLowerCase(), personNames.get(i).confidenceScore);
            }
            
            Tokenizer tokenizer = new Tokenizer();
            List<TokenGroup> tgInputList = tokenizer.getAllTokens(input, TOKEN_DELIM_1, MAX_TOKEN_COUNT);
            List<EntityTaggerOutput> organizationNames = organizationNameTagger.getOrganizationNames(input, tgInputList);
            Map<String, Double> organizationNamesConfidence = new LinkedHashMap<>();
            for (int i = 0; i < organizationNames.size(); ++i)
            {
                organizationNamesConfidence.put(organizationNames.get(i).tg.getToken().toLowerCase(), organizationNames.get(i).confidenceScore);
            }
            
            if (usePOSTagFiltering)
            {
                if (nouns == null)
                {
                    TokenGroup[] tokensForPOSTagging = getTokensForOpenNLPPOSTagging(input);
                    nouns = postagger.tag(tokensForPOSTagging, MAX_TOKEN_GROUP_LENGTH, ALLOW_PREPOSITIONS, ALLOW_CONJUNCTIONS, ALLOW_ADJECTIVES, ALLOW_DETERMINERS, ALLOW_ADVERBS);
                }
                nouns.forEach((keyTokenGroup, valuePOSTag)
                        -> 
                        {
                            String tokenLCase = keyTokenGroup.getToken().toLowerCase();

                            if (!unlikelyEntityNames.contains(tokenLCase))
                            {
                                if (!allMatchedStrings.contains(tokenLCase))
                                {
                                    if (!textUtil.hasOverlap(boundaries, keyTokenGroup.getIndex(), keyTokenGroup.getSize()))
                                    {
                                        if (!keyTokenGroup.getToken().contains(" "))
                                        {
                                            if (keyTokenGroup.getToken().length() > 1) // Discard single-character tokens
                                            {
                                                if (!wnutil.isInDictionary(tokenLCase)) // Discard single-word tokens which are found in an English language dictionary.
                                                {
                                                    unknownEntities.add(tokenLCase);

                                                    EntityTaggerOutput eto = new EntityTaggerOutput();
//                                                  eto.ek = new EntityKeyword(); // object created with new EntityKeyword instance.
                                                    eto.tg = keyTokenGroup;
                                                    if (eto.tg.getToken() != null)
                                                    {
//                                                        eto.tg.setToken(eto.tg.getToken().toLowerCase());
                                                        eto.tg.setToken(eto.tg.getToken());
                                                    }
                                                    eto.isDictionaryMatch = false;
//                                                    etoList.add(eto); // commented out adding single token nouns of unknown type
                                                    
                                                }
                                            }
                                        }
                                        else 
                                        {
//                                            if (keyTokenGroup.getToken().length() > 0 && digitsAndUpperCaseLetters.contains(keyTokenGroup.getToken().substring(0, 1)))
                                            if (keyTokenGroup.getToken().length() > 0)
                                            {
                                                unknownEntities.add(tokenLCase);

                                                EntityTaggerOutput eto = new EntityTaggerOutput();
                                                //                            eto.ek = new EntityKeyword(); // object created with new EntityKeyword instance.
                                                eto.tg = keyTokenGroup;
                                                if (eto.tg.getToken() != null)
                                                {
//                                                    eto.tg.setToken(eto.tg.getToken().toLowerCase());
                                                    eto.tg.setToken(eto.tg.getToken());
                                                }
                                                eto.isDictionaryMatch = false;
                                                
                                                // 
                                                Double confidence = null;
                                                Double personConfidence = personNamesConfidence.get(eto.tg.getToken().toLowerCase());
                                                Double organizationConfidence = organizationNamesConfidence.get(eto.tg.getToken().toLowerCase());
                                                if (personConfidence != null || organizationConfidence != null)
                                                {
                                                    if (personConfidence != null && organizationConfidence != null)
                                                    {
                                                        if (organizationConfidence >= personConfidence)
                                                        {
                                                            eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_ORGANIZATION;
                                                            confidence = organizationConfidence;
                                                        }
                                                        else
                                                        {
                                                            eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_PERSON;
                                                            confidence = personConfidence;
                                                        }                                                    
                                                    }
                                                    else if (personConfidence != null)
                                                    {
                                                        eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_PERSON;
                                                        confidence = personConfidence;                                                        
                                                    }
                                                    else
                                                    {
                                                        eto.entityType = net.infantai.nlp.ner.Constants.ENTITY_TYPE_ORGANIZATION;
                                                        confidence = organizationConfidence;                                                        
                                                    }
                                                    if (confidence != null)
                                                    {
                                                        if (!matches.containsKey(eto.tg))
                                                        {
                                                            eto.confidenceScore = confidence.doubleValue();
                                                            etoList.add(eto);                                                    
                                                        }
                                                    }                                                
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                });
            }
            else
            {
                // Nothing to do here, this condition has already been handled a 
                // few blocks above: search for "if (tagNonDictionaryEntities && !usePOSTagFiltering)"
            }
        }

        //
        t2 = System.nanoTime();
        LOGGER.debug("Tagging took: " + ((t2 - t1) / 1000000.0) + " ms. ");
        LOGGER.trace("===== Tagged Entities =====");
        matches.forEach((key, value)
                -> 
                {
                    LOGGER.trace(key + " => ");
                    value.forEach(ekw
                            -> 
                            {
                                LOGGER.trace("entity_type: " + ekw.getEntityTypeName() + " keyword_id: " + ekw.getKeywordId() + " entity_name: " + ekw.getEntityName() + " entity_id: " + ekw.getEntityId());
                    });
        });
        LOGGER.trace("===== Unknown Entities =====");
        unknownEntities.forEach((key)
                -> 
                {
                    LOGGER.trace(key);
        });
        
        Collections.sort(etoList); // EntityTaggerOutput objects will be tagged by index now.
        
        List<EntityTaggerOutput> etoListToReturn = combineNegationForAdjacentEntities(inputLCase, etoList, true);
        
        ValueExtractor ve = new ValueExtractor();
        Set<String> allowedPreTokens = new HashSet<>();
        allowedPreTokens.add("is");
        allowedPreTokens.add("was");
        Set<String> allowedPostTokens = new HashSet<>();
        allowedPostTokens.add("is");
        allowedPostTokens.add("of");
        allowedPostTokens.add("at");
        allowedPostTokens.add("is measured");
        allowedPostTokens.add("is measured at");
        allowedPostTokens.add("is measured to be");
        allowedPostTokens.add("is observed");
        allowedPostTokens.add("is observed at");
        allowedPostTokens.add("is observed to be");        
        allowedPostTokens.add("was");
        allowedPostTokens.add("was measured");
        allowedPostTokens.add("was measured at");
        allowedPostTokens.add("was measured to be");
        allowedPostTokens.add("was observed");
        allowedPostTokens.add("was observed at");
        allowedPostTokens.add("was observed to be");
        allowedPostTokens.add("is low");
        allowedPostTokens.add("is low today");
        allowedPostTokens.add("is little low today");
        allowedPostTokens.add("is a little low today");
        allowedPostTokens.add("is high");
        allowedPostTokens.add("is high today");
        allowedPostTokens.add("is little high today");
        allowedPostTokens.add("is a little high today");        
        allowedPostTokens.add("was low");
        allowedPostTokens.add("was low today");
        allowedPostTokens.add("was little low today");
        allowedPostTokens.add("was a little low today");
        allowedPostTokens.add("was high");
        allowedPostTokens.add("was high today");
        allowedPostTokens.add("was little high today");
        allowedPostTokens.add("was a little high today");                
        
        Set<Long> allowedEntityTypeIds = new HashSet<>();
        allowedEntityTypeIds.add(98L);
        allowedEntityTypeIds.add(99L);
        etoListToReturn = ve.mergeValuesAndModifyEntities(input, etoListToReturn, allowedPreTokens, allowedPostTokens, allowedEntityTypeIds);
/*
        LOG.trace("===== Date Entities =====");
        dateEsopList.forEach((esop)
                -> 
                {
                    LOG.trace("Date:" + esop.getEntityDateParts() + " keyword:" + esop.getEntityKeyword());
        });

        Collections.sort(etoList);

//        LOG.debug("Entity tagging output");  
//        for (int i = 0; i < etoList.size(); ++i)
//        {
//            EntityTaggerOutput eto = etoList.get(i);
//            LOG.debug("Entity ID = "+eto.ek.getEntityId() + "  Entity Keyword = " + eto.tg.getToken() + " Entity Name = " + eto.ek.getEntityName());
//        }

        List<SentimentOutput> sopList = new ArrayList<>();
        List<EntityTaggerOutput> etoListForSentimentComputation = new ArrayList<>();
        
        //exclude the entity type ids, which do not need sentiment tagging
        for (EntityTaggerOutput eto : etoList)
        {
            if (!entityTypeIdsToIgnoreForSentimentTagging.contains(eto.ek.getEntityTypeId()))
            {
                etoListForSentimentComputation.add(eto);
            }
        }
        
        //broker company context based sentiment tagging exclusion
        if(isBrokerTagged)
        {
            List<EntityTaggerOutput> etoListForSentimentComputationPostBrokerContextExclusion =  new ArrayList<>();
            if(isRatingsCategoryFoundInTitle && isTitle)
            {
                if(atLeastOneNonBrokerCompanyFound(etoList, brokerCompanyEntityIds))
                {
                    for (EntityTaggerOutput eto : etoListForSentimentComputation)
                    {
                        if (!brokerCompanyEntityIds.contains(eto.ek.getEntityId()))
                        {
                            etoListForSentimentComputationPostBrokerContextExclusion.add(eto);
                        }
                    }
                }
                else
                {
                    // either one or multiple broker companies without any exclusive company entities.
                    for (EntityTaggerOutput eto : etoListForSentimentComputation)
                    {
                        if (brokerCompanyEntityIds.contains(eto.ek.getEntityId()))
                        {
                            if (!appearsAsBroker(eto, input, brokerIdentificationPreTokens , brokerIdentificationPostTokens))
                            {
                                etoListForSentimentComputationPostBrokerContextExclusion.add(eto);
                            }
                        }
                        else
                        {
                            etoListForSentimentComputationPostBrokerContextExclusion.add(eto);
                        }
                    }
                
                }
            }
            else
            {
                //if rating is not tagged in the news or if its a description, check for the brokerage context tags
                for (EntityTaggerOutput eto : etoListForSentimentComputation)
                {
                    if (brokerCompanyEntityIds.contains(eto.ek.getEntityId()))
                    {
                        if (!appearsAsBroker(eto, input, brokerIdentificationPreTokens , brokerIdentificationPostTokens))
                        {
                            etoListForSentimentComputationPostBrokerContextExclusion.add(eto);
                        }
                    } 
                    else
                    {
                        etoListForSentimentComputationPostBrokerContextExclusion.add(eto);
                    }
                }
            }
            sopList = sentimentTagger.computeSentiment(input, etoListForSentimentComputationPostBrokerContextExclusion);
        }
        else
        {
            sopList = sentimentTagger.computeSentiment(input, etoListForSentimentComputation);
        }

        LOG.trace("Sentiment tags");
        for (int i = 0; i < sopList.size(); ++i)
        {
            SentimentOutput sop = sopList.get(i);
            LOG.trace("    " + sop.getEntityKeyword() + ": Sentiment Direction = " + sop.getSentiment() + " **** " + sop.getSentimentTerm());
        }
        esopList = computeDistinctEntitySentiment(sopList, etoList);
        computeConsolidatedSentiment(esopList);
        esopList.addAll(dateEsopList);
        LOG.debug("Computed Consolidated Entity Sentiment");
        for (int i = 0; i < esopList.size(); ++i)
        {
            EntitySentimentOutput esop = esopList.get(i);
            LOG.debug("ID:" + esop.getEntityId() + " Keyword:" + esop.getEntityKeyword() + " : Positive = " + esop.getSentimentPositive() + " Neutral " + esop.getSentimentNeutral() + " Negative = " + esop.getSentimentNegative() + " Consolidated = " + esop.getSentimentConsolidated());
        }
        return esopList;
        */

        return etoListToReturn;
    }
    
    private List<EntityTaggerOutput> combineNegationForAdjacentEntities(String input, List<EntityTaggerOutput> etoListSortedByIndex, boolean breakOnNewLine)
    {
        List<EntityTaggerOutput> etoList = new ArrayList<>();
        char[] charcatersToTrim = new char[] {',', ' ', '	'};
        char chCR = 13;
        char chLF = 10;
        for (int i = 0; i < etoListSortedByIndex.size(); ++i)
        {
            EntityTaggerOutput eto = etoListSortedByIndex.get(i);
            if (eto.tgNegation != null)
            {
                if (i < (etoListSortedByIndex.size() - 1)) // there's at least one more element
                {
                    EntityTaggerOutput etoNext = etoListSortedByIndex.get(i + 1);
                    int currentETOEndIndex = eto.tg.getIndex() + eto.tg.getSize();
                    if (currentETOEndIndex < etoNext.tg.getIndex())
                    {
                        String tib = input.substring(currentETOEndIndex, etoNext.tg.getIndex());
                        String tibTrimmed = tib.trim();
                        if (breakOnNewLine)
                        {
                            if (tib.indexOf(chCR) >= 0 || tib.indexOf(chLF) >= 0)
                            {
                                etoList.add(eto);
                                continue;
                            }
                        }
                        tibTrimmed = CommonUtil.trimCharacters(charcatersToTrim, tibTrimmed);
                        if (tibTrimmed.length() == 0 || tibTrimmed.equals("and") || tibTrimmed.equals("or") || tibTrimmed.equals("or even") || tibTrimmed.equals("nor")) // combine
                        {
                            etoNext.tgNegation = eto.tgNegation; 
                            LOGGER.debug("Carrying forward Negation from adjacent entities: " + eto.tg.getToken() + " {index " + eto.tg.getIndex() + "}" + " => " + etoNext.tg.getToken() + " {index " + etoNext.tg.getIndex() + "}");
                        }
                    }
                    else if (eto.tg.getIndex().intValue() == etoNext.tg.getIndex().intValue()) 
                    {
                        etoNext.tgNegation = eto.tgNegation; 
                        LOGGER.debug("Carrying forward Negation from overlapping entities: " + eto.tg.getToken() + " {index " + eto.tg.getIndex() + "}" + " => " + etoNext.tg.getToken() + " {index " + etoNext.tg.getIndex() + "}");                        
                    }
                }
            }
            etoList.add(eto);
        }
        return etoList;
    }

    /*
    private List<EntitySentimentOutput> computeDistinctEntitySentiment(List<SentimentOutput> sopList, List<EntityTaggerOutput> etoList)
    {
        Map<Long, EntitySentimentOutput> entityIdEsopMap = computeEntitySentimentForDictionaryEntities(sopList); // in case of multiple sentiment terms getting matched with the same entity, for each direction, pick the one with the highest weight
        Map<String, EntitySentimentOutput> unknownNounEsopMap = computeEntitySentimentForUnknownNouns(sopList); // in case of multiple sentiment terms getting matched with the same entity, for each direction, pick the one with the highest weight

        List<EntitySentimentOutput> finalSentimentOutput = new ArrayList<>();
        for (EntityTaggerOutput eto : etoList)
        {
            if (eto.isDictionaryMatch)
            {
                EntitySentimentOutput eso = entityIdEsopMap.get(eto.ek.getEntityId());
                if (eso == null)
                {
                    eso = new EntitySentimentOutput();
                    eso.setEntityId(eto.ek.getEntityId());
                    eso.setEntityKeywordId(eto.ek.getKeywordId());
                    eso.setEntityKeyword(eto.tg.getToken());
                    eso.setEntityKeywordIndex(eto.tg.getIndex());
                    entityIdEsopMap.put(eto.ek.getEntityId(), eso);
                }
            }
            else
            {
                EntitySentimentOutput eso = unknownNounEsopMap.get(eto.tg.getToken());
                if (eso == null)
                {
                    eso = new EntitySentimentOutput();
                    eso.setEntityId(eto.ek.getEntityId());
                    eso.setEntityKeywordId(eto.ek.getKeywordId());
                    eso.setEntityKeyword(eto.tg.getToken());
                    eso.setEntityKeywordIndex(eto.tg.getIndex());
                    unknownNounEsopMap.put(eto.tg.getToken(), eso);
                }
            }
        }
        finalSentimentOutput.addAll(entityIdEsopMap.values());
        finalSentimentOutput.addAll(unknownNounEsopMap.values());
        return finalSentimentOutput;
    }

    private Map<Long, EntitySentimentOutput> computeEntitySentimentForDictionaryEntities(List<SentimentOutput> sopList)
    {
        Map<Long, EntitySentimentOutput> esopMap = new HashMap<>();
        for (SentimentOutput sop : sopList)
        {
            if (sop.getEntityId() != -1)
            {
                EntitySentimentOutput esop = esopMap.get(sop.getEntityId());
                if (esop == null)
                {
                    esop = new EntitySentimentOutput();
                    esop.setEntityId(sop.getEntityId());
                }
                if (esop.getEntityKeywordIndex() == -1 || sop.getEntityKeywordIndex() < esop.getEntityKeywordIndex())
                {
                    esop.setEntityKeywordId(sop.getEntityKeywordId());
                    esop.setEntityKeyword(sop.getEntityKeyword());
                    esop.setEntityKeywordIndex(sop.getEntityKeywordIndex());
                }
                if (sop.getSentiment() == 1 && esop.getSentimentPositive() < sop.getWeight())
                {
                    esop.setSentimentPositive(sop.getWeight());
                    esop.setSentimentTermIdPositive(sop.getSentimentTermId());
                }
                else
                {
                    if (sop.getSentiment() == 0 && esop.getSentimentNeutral() < sop.getWeight())
                    {
                        esop.setSentimentNeutral(sop.getWeight());
                        esop.setSentimentTermIdNeutral(sop.getSentimentTermId());
                    }
                    else
                    {
                        if (sop.getSentiment() == -1 && esop.getSentimentNegative() < sop.getWeight())
                        {
                            esop.setSentimentNegative(sop.getWeight());
                            esop.setSentimentTermIdNegative(sop.getSentimentTermId());
                        }
                    }
                }
                esopMap.put(esop.getEntityId(), esop);
            }
        }

        return esopMap;
    }

    private Map<String, EntitySentimentOutput> computeEntitySentimentForUnknownNouns(List<SentimentOutput> sopList)
    {
        Map<String, EntitySentimentOutput> esopNounMap = new HashMap<>();
        for (SentimentOutput sop : sopList)
        {
            if (sop.getEntityId() == -1)
            {
                EntitySentimentOutput esop = esopNounMap.get(sop.getEntityKeyword());
                if (esop == null)
                {
                    esop = new EntitySentimentOutput();
                    esop.setEntityId(sop.getEntityId());
                    esop.setEntityKeywordId(sop.getEntityKeywordId());
                    esop.setEntityKeyword(sop.getEntityKeyword());
                    esop.setEntityKeywordIndex(sop.getEntityKeywordIndex());
                }
                if (sop.getSentiment() == 1 && esop.getSentimentPositive() < sop.getWeight())
                {
                    esop.setSentimentPositive(sop.getWeight());
                    esop.setSentimentTermIdPositive(sop.getSentimentTermId());
                    esop.setEntityKeywordId(sop.getEntityKeywordId());
                }
                else
                {
                    if (sop.getSentiment() == 0 && esop.getSentimentNeutral() < sop.getWeight())
                    {
                        esop.setSentimentNeutral(sop.getWeight());
                        esop.setSentimentTermIdNeutral(sop.getSentimentTermId());
                    }
                    else
                    {
                        if (sop.getSentiment() == -1 && esop.getSentimentNegative() < sop.getWeight())
                        {
                            esop.setSentimentNegative(sop.getWeight());
                            esop.setSentimentTermIdNegative(sop.getSentimentTermId());
                        }
                    }
                }
                esopNounMap.put(sop.getEntityKeyword(), esop);
            }
        }
        return esopNounMap;
    }

    private void computeConsolidatedSentiment(List<EntitySentimentOutput> esopList)
    {
        for (EntitySentimentOutput esop : esopList)
        {
            double sumOfSentimentWeights = esop.getSentimentNegative() + esop.getSentimentNeutral() + esop.getSentimentPositive();

            if (sumOfSentimentWeights > 0)
            {
                double sentimentNegative = esop.getSentimentNegative() / sumOfSentimentWeights;
                double sentimentNeutral = esop.getSentimentNeutral() / sumOfSentimentWeights;
                double sentimentPositive = esop.getSentimentPositive() / sumOfSentimentWeights;

                double sentimentConsolidated = 0.0;

                int positiveSentimentMaxWeight = sentimentTagger.getMaxPositiveWeight();
                int negativeSentimentMaxWeight = sentimentTagger.getMaxNegativeWeight();

                sentimentConsolidated = (sentimentPositive * (esop.getSentimentPositive() / positiveSentimentMaxWeight)) - (sentimentNegative * (esop.getSentimentNegative() / negativeSentimentMaxWeight));

                esop.setSentimentNegative(sentimentNegative);
                esop.setSentimentNeutral(sentimentNeutral);
                esop.setSentimentPositive(sentimentPositive);
                esop.setSentimentConsolidated(sentimentConsolidated);
                LOG.trace("Consolidated Sentiment is:" + sentimentConsolidated);
            }

        }

    }
    */

    /**
     * Gets tokens as Strings from TokenGroup objects. Converts to lower case.
     *
     * @param distinctTokens
     * @return
     */
    public static Set<String> getAllTokensForBTRM(Set<TokenGroup> distinctTokens)
    {
        Set<String> articleTokens = new HashSet<String>();

        Iterator<TokenGroup> iterArticleTokens = distinctTokens.iterator();
        while (iterArticleTokens.hasNext())
        {
            String strToken = iterArticleTokens.next().getToken();
            articleTokens.add(strToken.toLowerCase());
        }
        return articleTokens;
    }

//    private boolean appearsAsBroker(EntityTaggerOutput eto, String input, Set<String> brokerIdentificationPreTokens , Set<String> brokerIdentificationPostTokens)
//    {
//        boolean conclusivelyTaggedAsBroker = false;
//        try
//        {
//            int tindex = eto.tg.getIndex();
//            TokenGroup tg = eto.tg;
//            String titleLCase = input.toLowerCase();
//
//            if (tindex < titleLCase.length())
//            {
//                int lindex = tg.getIndex();
//                int rindex = tg.getIndex() + tg.getToken().length();
//                if (rindex >= titleLCase.length() || titleLCase.substring((rindex + 1), titleLCase.length()).trim().equals("analyst")) // check if this is the last token in the title
//                {
//                    if (lindex > 0)
//                    {
//                        char lchar = titleLCase.charAt(lindex - 1);
//                        if (lchar == ':' || lchar == '-' || lchar == '?')
//                        {
//                            conclusivelyTaggedAsBroker = true;
//                        }
//
//                        if (!conclusivelyTaggedAsBroker)
//                        {
//                            if (lindex > 1)
//                            {
//                                if (lchar == ' ')
//                                {
//                                    char lchar2 = titleLCase.charAt(lindex - 2);
//                                    if (lchar2 == ':' || lchar2 == '-' || lchar2 == '?')
//                                    {
//                                        conclusivelyTaggedAsBroker = true;
//                                    }
//                                }
//                            }
//                        }
//                    }
//
////                    if (!conclusivelyTaggedAsBroker)
////                    {
////
////                        String prevToken = "";
////                        if (lindex > 0)
////                        {
////                            prevToken = sentimentTagger.getNTokensFromPrevText(titleLCase.substring(0, lindex), 1);
////                            if (brokerVerbsToCheck.contains(prevToken))
////                            {
////                                conclusivelyTaggedAsBroker = true;
////                            }
////
////                        }
////                    }
//
//                }
//                
//                if (!conclusivelyTaggedAsBroker)
//                {
//                    if ((rindex + 1) < titleLCase.length())
//                    {
//                        Set<String> nextTokens = textUtil.getUpToNTokensFromNextText(titleLCase.substring(rindex + 1), 4);
//                        for(String nextToken : nextTokens)
//                        {
//                            if (brokerIdentificationPostTokens.contains(nextToken)) // contextual reference to a broker, not a company
//                            {
//                                conclusivelyTaggedAsBroker = true;
//                                break;
//                            }
//                        }
//                    }
//                }
//                if (!conclusivelyTaggedAsBroker)
//                {
//                    if (lindex > 0)
//                    {
//                        Set<String> prevTokens = textUtil.getUpToNTokensFromPrevText(titleLCase.substring(0, lindex), 4);
//                        for(String prevToken : prevTokens)
//                        {
//                            if (brokerIdentificationPreTokens.contains(prevToken))
//                            {
//                                conclusivelyTaggedAsBroker = true;
//                                break;
//                            }
//                        }
//                    }
//                }
//            }
//        }
//        catch (Exception ex)
//        {
//            LOG.error(ex.getMessage(), ex);
//        }
//
//        return conclusivelyTaggedAsBroker;
//    }

//    private boolean atLeastOneNonBrokerCompanyFound(List<EntityTaggerOutput> etoList, Set<Long> brokerCompanyEntityIds)
//    {
//        boolean atLeastOneNonBrokerCompanyFound = false;
//        for(EntityTaggerOutput eto: etoList)
//        {
//            if(eto.ek.getEntityTypeId()== ENTITY_TYPE_COMPANY)
//            {
//                if(!brokerCompanyEntityIds.contains(eto.ek.getEntityId()))
//                {
//                    atLeastOneNonBrokerCompanyFound = true;
//                    break;
//                }
//            }
//        }
//        return atLeastOneNonBrokerCompanyFound;
//    }
}
