/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package net.infantai.util.common;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.GregorianCalendar;
import java.util.List;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class BooleanExpressionParser 
{
    
    private static final String BOOLEAN_OPERATOR_LOGICAL_AND = "&&";
    private static final String BOOLEAN_OPERATOR_LOGICAL_OR = "||";
    private static final String BOOLEAN_OPERAND_TRUE = "true";
    private static final String BOOLEAN_OPERAND_NOT_FALSE = "!false";
    private static final String BOOLEAN_OPERAND_FALSE = "false";
    private static final String BOOLEAN_OPERAND_NOT_TRUE = "!true";    

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) 
    {
        String expr = "true || (true && false) && false || ((true||false) && (!false || true))";
        try
        {
            System.out.println("Result: " + parseExpression(expr));
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
    }
    
    public static boolean parseExpression(String expr) throws MalformedExpressionException
    {
        return evaluateSimpleExpression(reduceBrackets(expr));
    }
    
    /**
     * Recursive function to reduce brackets.
     * @param expr
     * @return
     * @throws MalformedExpressionException 
     */
    private static String reduceBrackets(String expr) throws MalformedExpressionException
    {
        String reducedExpr = expr;
        Deque<Integer> leftBracketIndexStack = new ArrayDeque<Integer>(); // Using ArrayDeque (Double-Ended Queue) (instead of ArrayList) as a Stack, for slightly better performance.
        boolean foundLBracket = false;
        
        int i = 0;
        for (i = 0; i < expr.length(); ++i)
        {
            if (expr.charAt(i) == '(')
            {
                leftBracketIndexStack.push(new Integer(i));
                foundLBracket = true;
            }
            else if (expr.charAt(i) == ')')
            {
                if (leftBracketIndexStack.isEmpty())
                {
                    throw new MalformedExpressionException("Missing '(' In Sub-Expression !" /*+ expr*/);
                }
                String subexpr = expr.substring(leftBracketIndexStack.getFirst() + 1, i);
                reducedExpr = expr.substring(0, leftBracketIndexStack.getFirst());
                
                boolean seres = evaluateSimpleExpression(subexpr);
                if (i <= (expr.length() - 1))
                {
                    reducedExpr = reducedExpr + ' ' + seres + ' ' + expr.substring(i + 1, expr.length());
                }
                
                leftBracketIndexStack.pop();
//                System.out.println("Subexpr => " + subexpr + " Reduced = " + reducedExpr);
                break;
            }
        }
        
        if (foundLBracket)
        {
            if (i == expr.length())
            {
                throw new MalformedExpressionException("Missing ')' In Sub-Expression !" /*+ expr*/);
            }
            return reduceBrackets(reducedExpr);
        }
        return reducedExpr;
    }    
    
    /**
     * Recursive function to evaluate simple boolean expressions. 
     * @param subexpr
     * @return
     * @throws MalformedExpressionException
     * @throws NullPointerException 
     */
    private static boolean evaluateSimpleExpression(String subexpr) throws MalformedExpressionException, NullPointerException
    {
//        System.out.println("Simple Expression => " + subexpr);
        if (subexpr == null)
        {
            throw new NullPointerException("Sub-Expression NULL !");
        }
        subexpr = subexpr.replace("! ", "!");
        Tokenizer st = new Tokenizer();
        String exprToEval = subexpr.trim();
        List<TokenGroup> tokens = st.getAllTokens(exprToEval, " &|", 1);
        boolean b1 = false;
        boolean b2 = false;
        boolean res = false;
        if (tokens.size() > 1)
        {
            TokenGroup tg1 = tokens.get(0);
            TokenGroup tg2 = tokens.get(1);
            String token1 = tg1.getToken();
            String token2 = tg2.getToken();
            if (token1.equals(BOOLEAN_OPERAND_TRUE) || token1.equals(BOOLEAN_OPERAND_NOT_FALSE))
            {
                b1 = true;
            }
            else if (token1.equals(BOOLEAN_OPERAND_FALSE) || token1.equals(BOOLEAN_OPERAND_NOT_TRUE))
            {
                b1 = false;
            }
            else
            {
                throw new MalformedExpressionException("Unacceptable Boolean Operand => " + token1);
            }
            
            if (token2.equals(BOOLEAN_OPERAND_TRUE) || token2.equals(BOOLEAN_OPERAND_NOT_FALSE))
            {
                b2 = true;
            }
            else if (token2.equals(BOOLEAN_OPERAND_FALSE) || token2.equals(BOOLEAN_OPERAND_NOT_TRUE))
            {
                b2 = false;
            }
            else
            {
                throw new MalformedExpressionException("Unacceptable Boolean Operand => " + token2);
            }
            
            String operator = exprToEval.substring((tg1.getIndex() + token1.length()), tg2.getIndex()).trim();
//            System.out.println("Operator => " + operator);
            if (operator.equals(BOOLEAN_OPERATOR_LOGICAL_AND))
            {
                res = b1 && b2;
            }
            else if (operator.equals(BOOLEAN_OPERATOR_LOGICAL_OR))
            {
                res = b1 || b2;
            }
            else
            {
                throw new MalformedExpressionException("Unknown Operator => " + operator);
            }
            
            String strRes = BOOLEAN_OPERAND_FALSE;
            if (res)
            {
                strRes = BOOLEAN_OPERAND_TRUE;
            }
//            System.out.println("Reduced Simple Expression => " + strRes + exprToEval.substring(tg2.getIndex() + token2.length()));
            return evaluateSimpleExpression(strRes + exprToEval.substring(tg2.getIndex() + token2.length()));
        }
        else if (tokens.size() == 1)
        {
            if (tokens.get(0).getToken().equals(BOOLEAN_OPERAND_TRUE) || tokens.get(0).getToken().equals(BOOLEAN_OPERAND_NOT_FALSE))
            {
                res = true;
            }
            else if (tokens.get(0).getToken().equals(BOOLEAN_OPERAND_FALSE) || tokens.get(0).getToken().equals(BOOLEAN_OPERAND_NOT_TRUE))
            {
                res = false;
            }
            else
            {
                throw new MalformedExpressionException("Unacceptable Boolean Operand => " + tokens.get(0).getToken());
            }            
        }
        else
        {
            throw new MalformedExpressionException("Sub-Expression Empty !");
        }
        return res;
    }
    
}
