/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * FileIO.java
 *
 * Created on 27 July 2008, 21:08
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package net.infantai.util.common;

import java.io.*;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 *
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class FileIO
{
    
    /**
     * Creates a new instance of FileIO.
     */
    public FileIO() 
    {
        
    }
    
    /**
     * Writes the specified byte array to a disk file. 
     */
    public static boolean writeBinaryFile(String filePath, byte[] fileData, boolean overwrite) throws FileNotFoundException, IOException, SecurityException
    {
        boolean fileWritten = false;
        File file = new File(filePath);
        FileOutputStream fos = null;
        try
        {
            boolean fileExists = file.exists();
            if (fileExists)
            {
                if (!overwrite)
                {
                    return fileWritten;
                }
            }
            else
            {
                boolean fileCreated = file.createNewFile();
                if (!fileCreated)
                {
                    return fileWritten;
                }
            }
            if (!file.isFile() || !file.canWrite())
            {
                return fileWritten;
            }            
            fos = new FileOutputStream(file);
            fos.write(fileData);
            fileWritten = true;
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
            throw fnfe;
        }
        catch (IOException ioe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
            throw ioe;
        }
        catch (SecurityException se)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + se.getMessage());
            throw se;
        }        
        finally
        {
            if (fos != null)
            {
                try
                {
                    fos.close();
                }
                catch (IOException ioe)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                    ioe.printStackTrace();
                }
            }
        }
        return fileWritten;
    }

    /**
     * Reads the contents of the specified file into a byte array.
     */
    public static byte[] readBinaryFile(String filePath) throws UnsupportedFileSizeException, FileNotFoundException, IOException
    {
        byte[] fileData = null;
        FileInputStream fis = null;
        int offset = 0;
        int numRead = 0;
        try
        {
            File file = new File(filePath);
            fis = new FileInputStream(file);
            long fileSize = file.length();
            if (fileSize >= Integer.MAX_VALUE)
            {
                throw new UnsupportedFileSizeException("File Size = " + ((double) fileSize / (1024 * 1024)) + " MiB");
            }
            int iFileSize = (int) fileSize;
            fileData = new byte[iFileSize];
            while ((offset < fileData.length) && ((numRead = fis.read(fileData, offset, (fileData.length - offset))) >= 0))
            {
                offset += numRead;
            }
        }
        catch (FileNotFoundException fnfe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + fnfe.getMessage());
            throw fnfe;
        }
        catch (IOException ioe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
            throw ioe;
        }
        finally
        {
            if (fis != null)
            {
                try
                {
                    fis.close();
                }
                catch (IOException ioe)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                    ioe.printStackTrace();
                }
            }
        }
        return fileData;
    }
    
    /**
     * Writes the specified input String to a disk file.
     */
    public static boolean writeTextFile(String filePath, String fileData, boolean overwrite) throws IOException
    {
        boolean fileWritten = false;
        File file = new File(filePath);
        Writer writer = null;
        try
        {
            boolean fileExists = file.exists();
            if (fileExists)
            {
                if (!overwrite)
                {
                    return fileWritten;
                }
            }
            else
            {
                boolean fileCreated = file.createNewFile();
                if (!fileCreated)
                {
                    return fileWritten;
                }
            }
            if (!file.isFile() || !file.canWrite())
            {
                return fileWritten;
            }
            writer = new BufferedWriter(new FileWriter(file));
            writer.write(fileData);
            fileWritten = true;
        }
        catch (IOException ioe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
            throw ioe;
        }
        finally
        {
            if (writer != null)
            {
                try
                {
                    writer.close();
                }
                catch (IOException ioe)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                    ioe.printStackTrace();
                }
            }
        }
        return fileWritten;
    }
    
    /**
     * Reads the contents of the specified file into a String; 
     * replaces all newline characters in the input file with spaces.
     */
    public static String readTextFile(String filePath) throws IOException
    {
        StringBuilder fileData = new StringBuilder();
        String strFileData = null;
        try
        {
            BufferedReader brinput = new BufferedReader(new FileReader(new File(filePath)));
            String strLine = null;
            try
            {
                while ((strLine = brinput.readLine()) != null)
                {
                    fileData.append(strLine);
                    fileData.append(" ");
                }            
            }
            catch (IOException ioe)
            {
                System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
                throw ioe;
            }
            finally
            {
                brinput.close();
            }
            strFileData = fileData.toString().trim();
        }
        catch (IOException ioe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
            throw ioe;
        }
        return strFileData;
    }
    
    /**
     * Reads the contents of the specified file and returns an ArrayList.
     * 
     */
    public static List<String> readTextFileAsList(String filePath) throws IOException
    {
        List<String> fileLines = null;
        BufferedReader brinput = null;
        try
        {
            brinput = new BufferedReader(new FileReader(new File(filePath)));
            fileLines = new ArrayList<String>();
            String strLine = null;
            while ((strLine = brinput.readLine()) != null)
            {
                fileLines.add(strLine);
            }
//            brinput.close();
        }
        catch (IOException ioe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
            throw ioe;
        }
        finally
        {
            if (brinput != null)
            {
                brinput.close();
            }
        }
        return fileLines;
    }    
    
    /**
     * Reads the contents of the specified file and returns a LinkedHashSet.
     * Also converts all text to lower case.
     */
    public static Set<String> readTextFileAsSet(String filePath) throws IOException
    {
        Set<String> fileLines = null;
        BufferedReader brinput = null;
        FileReader fr = null;
        try
        {
            fr = new FileReader(new File(filePath));
            brinput = new BufferedReader(fr);
            fileLines = new LinkedHashSet<String>();
            String strLine = null;
            while ((strLine = brinput.readLine()) != null)
            {
                fileLines.add(strLine.toLowerCase());
            }
//            brinput.close();
        }
        catch (IOException ioe)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioe.getMessage());
            throw ioe;
        }
        finally
        {
            if (brinput != null)
            {
                try
                {
                    brinput.close();
                }
                catch (IOException ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                }
            }
            if (fr != null)
            {
                try
                {
                    fr.close();
                }
                catch (Exception ioex)
                {
                    System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ioex.getMessage());
                }
            }
        }
        return fileLines;
    }        
    
    /**
     * Reads the contents of the specified URL into a byte array. Returns null if an exception occurs.
     */
    public static byte[] readFileFromURL(String strFileURL, String userAgent, long maxFileSize, boolean followRedirects)
    {
        byte[] fileBytes = null;
        FetchedFile ff = null;
        try
        {
            ff = CommonUtil.fetch(strFileURL, userAgent, maxFileSize, followRedirects);
            if (ff != null)
            {
                fileBytes = ff.getFileBytes();
            }            
        }
        catch (Exception ex)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + ex.getMessage());
            ex.printStackTrace();
        }
        return fileBytes;
    }
    
}
