/*
 * This Java source file is completely Free and Open-Source, 
 * released under the MIT License.
 *
 * Copyright (2012) Ashish Nawal Saharia (ashish.saharia@yahoo.co.in)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 * OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

/*
 * HashUtil.java
 *
 * Created on 05 May 2008, 18:02
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package net.infantai.util.common;


import java.io.UnsupportedEncodingException;
import java.security.*;
import java.util.GregorianCalendar;
import java.util.zip.Adler32;
import java.util.zip.CRC32;


/**
 * This class provides methods to compute Checksums and Cryptographic 
 * Hashes (Message Digests). The following algorithms are currently supported:
 * <br>
 * <ul>
 * <li>CRC-32 (HashUtil.CHECKSUM_ALGORITHM_CRC32)</li>
 * <li>ADLER-32 (HashUtil.CHECKSUM_ALGORITHM_ADLER32)</li>
 * <li>MD5 (HashUtil.MESSAGE_DIGEST_ALGORITHM_MD5)</li>
 * <li>SHA-1 (HashUtil.MESSAGE_DIGEST_ALGORITHM_SHA1)</li>
 * <li>SHA-256 (HashUtil.MESSAGE_DIGEST_ALGORITHM_SHA256)</li>
 * <li>SHA-384 (HashUtil.MESSAGE_DIGEST_ALGORITHM_SHA384)</li>
 * <li>SHA-512 (HashUtil.MESSAGE_DIGEST_ALGORITHM_SHA512)</li>
 * </ul>
 * 
 * @author Ashish (ashish.saharia@yahoo.co.in)
 */
public class HashUtil 
{
    
    private MessageDigest md = null;
    private CRC32 crc32 = null;
    private Adler32 adler32 = null;
    
    public static final String CHECKSUM_ALGORITHM_CRC32 = "CRC-32";
    public static final String CHECKSUM_ALGORITHM_ADLER32 = "ADLER-32";
    public static final String MESSAGE_DIGEST_ALGORITHM_MD5 = "MD5";
    public static final String MESSAGE_DIGEST_ALGORITHM_SHA1 = "SHA-1";
    public static final String MESSAGE_DIGEST_ALGORITHM_SHA256 = "SHA-256";
    public static final String MESSAGE_DIGEST_ALGORITHM_SHA384 = "SHA-384";
    public static final String MESSAGE_DIGEST_ALGORITHM_SHA512 = "SHA-512";
    
    /**
     * Creates a new instance of HashUtil. 
     */    
    public HashUtil(String algorithmName) 
    {
        try
        {
            if (algorithmName.equalsIgnoreCase(CHECKSUM_ALGORITHM_CRC32))
            {
                crc32 = new CRC32();
            }
            else if (algorithmName.equalsIgnoreCase(CHECKSUM_ALGORITHM_ADLER32))
            {
                adler32 = new Adler32();
            }
            else
            {
                md = MessageDigest.getInstance(algorithmName);
            }
        }
        catch(NoSuchAlgorithmException nsae)
        {
            System.err.println(new GregorianCalendar().toZonedDateTime() + " => " + nsae.getMessage());
            nsae.printStackTrace();
        }
    }
    

    /**
     * Returns a hash (encoded as a lower-case hexadecimal string) of the input 
     * byte array. 
     * @param inputData Input data to be hashed.
     * @return The hash (encoded as a lower-case hexadecimal string) of the input byte array.
     */
    public String getHash(byte[] inputData)
    {
        String strHash = null;
        byte[] hash = getHashRaw(inputData);
        if (hash != null)
        {
            return ByteInterpreter.bytesToHexString(hash, false);
        }
        return strHash;
    }
    
    /**
     * Returns a hash of the input byte array. 
     * Note that this number can be negative, since Java does not natively 
     * support unsigned integers. 
     * @param inputData Input data to be hashed.
     * @return A Long value representing the hash of the input byte array.
     * @throws UnsupportedOperationException
     * @throws ByteInterpretationException 
     */    
    public Long getHashNumeric(byte[] inputData) throws UnsupportedOperationException, ByteInterpretationException
    {
        Long hashNumeric = null;
        byte[] hash = getHashRaw(inputData);
        if (hash != null)
        {
            int hashBitLength = hash.length * 8;
            if (hashBitLength > 64)
            {
                throw new UnsupportedOperationException("Can't encode " + (hashBitLength * 8) + " bits as a Long value.");
            }
            hashNumeric = new Long(ByteInterpreter.bytesToLong(hash, 0, ByteInterpreter.BYTE_ORDER_BIG_ENDIAN));
        }
        return hashNumeric;
    }    
    
    /**
     * Returns a byte array representing a hash of the input data. The returned 
     * byte array can subsequently be encoded as a hexadecimal string, or as an 
     * integer / long value. 
     * @param inputData Input data to be hashed.
     * @return A byte array representing the hash of the input data.
     */
    public byte[] getHashRaw(byte[] inputData)
    {
        byte[] hash = null;
        if (md != null)
        {
            md.update(inputData);
            hash = md.digest();
            md.reset();
        }
        else if (adler32 != null)
        {
            adler32.update(inputData);
            hash = ByteConverter.toByteArray((int) adler32.getValue());
            adler32.reset();
        }        
        else if (crc32 != null)
        {
            crc32.update(inputData);
            hash = ByteConverter.toByteArray((int) crc32.getValue());
            crc32.reset();
        }
        return hash;
    }    
    

    /**
     * Convenience method. Returns a hash of the input String. Note that this 
     * method uses UTF-8 to encode the input String as an array of bytes.
     * @param input Input data to be hashed.
     * @return The hash (encoded as a lower-case hexadecimal string) of the input String.
     */
    public String getHash(String input)
    {
        byte[] inputBytes;
        inputBytes = input.getBytes(java.nio.charset.StandardCharsets.UTF_8);
        return getHash(inputBytes);
    }
    
    /**
     * Returns a hash of the input String.
     * @param input Input data to be hashed.
     * @param inputEncoding The character encoding (like "US-ASCII", "ISO-8859-1", "UTF-8", etc.) 
     * to use, for encoding the input String as an array of bytes, before computing the hash. 
     * @return The hash (encoded as a lower-case hexadecimal string) of the input String.
     * @throws UnsupportedEncodingException 
     */
    public String getHash(String input, String inputEncoding) throws UnsupportedEncodingException
    {
        byte[] inputBytes;
        inputBytes = input.getBytes(inputEncoding);
        return getHash(inputBytes);
    }    
    
    
    public static void main(String[] args) throws UnsupportedEncodingException
    {
        HashUtil crc32 = new HashUtil(CHECKSUM_ALGORITHM_CRC32);
        HashUtil adler32 = new HashUtil(CHECKSUM_ALGORITHM_ADLER32);
        HashUtil md5 = new HashUtil(MESSAGE_DIGEST_ALGORITHM_MD5);
        HashUtil sha1 = new HashUtil(MESSAGE_DIGEST_ALGORITHM_SHA1);
        HashUtil sha256 = new HashUtil(MESSAGE_DIGEST_ALGORITHM_SHA256);
        HashUtil sha384 = new HashUtil(MESSAGE_DIGEST_ALGORITHM_SHA384);
        HashUtil sha512 = new HashUtil(MESSAGE_DIGEST_ALGORITHM_SHA512);
        
        String input = "Pack my box with five dozen liquor jugs.";
        System.out.println(CHECKSUM_ALGORITHM_CRC32 + ": " +  crc32.getHash(input));
        System.out.println(CHECKSUM_ALGORITHM_ADLER32 + ": " + adler32.getHash(input));
        System.out.println(MESSAGE_DIGEST_ALGORITHM_MD5 + ": " + md5.getHash(input));
        System.out.println(MESSAGE_DIGEST_ALGORITHM_SHA1 + ": " + sha1.getHash(input));
        System.out.println(MESSAGE_DIGEST_ALGORITHM_SHA256 + ": " + sha256.getHash(input));
        System.out.println(MESSAGE_DIGEST_ALGORITHM_SHA384 + ": " + sha384.getHash(input));
        System.out.println(MESSAGE_DIGEST_ALGORITHM_SHA512 + ": " + sha512.getHash(input));

    }
    
}
